'''
Creates an equilibrium before saving it as a geqdsk file.
Then reconstructs the equilibrium from the geqdsk using
constraints on coil currents.
'''

from freegs import geqdsk
from freegs import machine
from freegs.plotting import plotEquilibrium

import freegs

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.TestTokamak()

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-1.0, Zmax=1.0,   # Height range
                        nx=65, ny=65,          # Number of grid points
                        boundary=freegs.boundary.freeBoundaryHagenow)  # Boundary condition

#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        1e3, # Plasma pressure on axis [Pascals]
                                        2e5, # Plasma current [Amps]
                                        2.0) # Vacuum f=R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(1.1, -0.6),   # (R,Z) locations of X-points
           (1.1, 0.8)]

isoflux = [(1.1,-0.6, 1.1,0.6)] # (R1,Z1, R2,Z2) pair of locations

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The toroidal current profile function
             constrain,
             show=True)   # Constraint function to set coil currents

# eq now contains the solution

# Currents in the coils
tokamak.printCurrents()

psi1 = eq.psi()

#########################################

# Save to G-EQDSK file

from freegs import geqdsk

with open("lsn.geqdsk", "w+") as f:
    geqdsk.write(eq, f)

#########################################
# Read in the geqdsk file with bounds on the coil currents

# Lower/Upper limits of coil currents
current_lims = [(140000.0,155000.0),(60000.0,63000.0),(-105000.0,-90000.0),(-60000.0,-55000.0)]

tokamak2 = freegs.machine.TestTokamak()

with open("lsn.geqdsk") as f:
    eq2= geqdsk.read(f, tokamak2, show=True, current_bounds=current_lims)

# eq2 now contains the solution

plotEquilibrium(eq2)

# Currents in the coils
tokamak2.printCurrents()

# Compare reconstructed psi v original psi
psi2 = eq2.psi()

pct_change = abs(100.0*(psi2 - psi1)/psi1)

import numpy as np
import matplotlib.pyplot as plt

fig, ax = plt.subplots()

ax.imshow(pct_change.T, extent=[min(eq.R[:, 0]),max(eq.R[:, 0]),min(eq.Z[0, :]),max(eq.Z[0, :])],
        origin='lower', vmax=20.0)

ax.plot(eq.tokamak.wall.R,eq.tokamak.wall.Z,'k')
ax.contour(eq.R,eq.Z,eq.psi(),levels=[eq.psi_bndry],colors='w')
ax.contour(eq2.R,eq2.Z,eq2.psi(),levels=[eq2.psi_bndry],colors='r')
ax.plot([],[],'r',label='original')
ax.plot([],[],'w',label='achieved')

ax.set_xlabel('R(m)')
ax.set_ylabel('Z(m)')
ax.set_aspect('equal')
ax.set_title(r'pct diff in $\psi$')
ax.legend()

plt.show()



# -*- coding: utf-8 -*-
#
# FreeGS documentation build configuration file, created by
# sphinx-quickstart on Thu May 18 23:46:05 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from importlib.metadata import version as get_version

sys.path.insert(0, os.path.abspath(".."))

from unittest.mock import MagicMock


class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


MOCK_MODULES = ["_tkinter", "shapely", "shapely.geometry", "shapely.geometry.polygon"]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "FreeGS"
copyright = "2017-2023, FreeGS authors"
author = "Ben Dudson"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags
release = get_version(project)
# Major.minor version
version = ".".join(release.split(".")[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Numpy-doc config
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_attr_annotations = True
napoleon_preprocess_types = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "repository_url": "https://github.com/freegs-plasma/freegs",
    "repository_branch": "main",
    "path_to_docs": "docs",
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_issues_button": True,
    "home_page_in_toc": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "FreeGSdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "FreeGS.tex", "FreeGS Documentation", "Ben Dudson", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "freegs", "FreeGS Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "FreeGS",
        "FreeGS Documentation",
        author,
        "FreeGS",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# -- Options for intersphinx extension ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "xarray": ("https://docs.xarray.dev/en/latest", None),
    "pint": ("https://pint.readthedocs.io/en/stable/", None),
}



#
# Performs the same calculation with two different boundary locations
# and compares the result


import freegs

from numpy import linspace, amin, amax, meshgrid, exp

import matplotlib.pyplot as plt

tokamak = freegs.machine.TestTokamak()

eq1 = freegs.Equilibrium(tokamak=tokamak,
                         Rmin=0.1, Rmax=2.0,    # Radial domain
                         Zmin = -2.0, Zmax = 2.0,
                         nx=65, ny=129)          # Number of grid points

tokamak = freegs.machine.TestTokamak()

eq2 = freegs.Equilibrium(tokamak=tokamak,
                         Rmin=0.1, Rmax=2.0,    # Radial domain
                         Zmin=-1.0, Zmax=1.0,   # Height range
                         nx=65, ny=65)          # Number of grid points

# Set the initial psi

xx, yy = meshgrid(linspace(eq1.Rmin,eq1.Rmax,65), linspace(eq1.Zmin,eq1.Zmax,129), indexing='ij')
psi = exp( - ( (xx - 1.0)**2 + (yy)**2 ) / 0.3**2 )
eq1._updatePlasmaPsi(psi)

xx, yy = meshgrid(linspace(eq2.Rmin,eq2.Rmax,65), linspace(eq2.Zmin,eq2.Zmax,65), indexing='ij')
psi = exp( - ( (xx - 1.0)**2 + (yy)**2 ) / 0.3**2 )
eq2._updatePlasmaPsi(psi)


xpoints = [(1.1, -0.6),   # (R,Z) locations of X-points
           (1.1, 0.8)]

isoflux = [(1.1,-0.6, 1.1,0.6)]

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

constrain(eq1)
constrain(eq2)

profiles1 = freegs.jtor.ConstrainBetapIp(eq1, 0.1, 1e6, 1.0)
profiles2 = freegs.jtor.ConstrainBetapIp(eq2, 0.1, 1e6, 1.0)

jtor1 = profiles1.Jtor(eq1.R, eq1.Z, eq1.psi(), eq1.psi_bndry)
jtor2 = profiles2.Jtor(eq2.R, eq2.Z, eq2.psi(), eq2.psi_bndry)

# Check jtor1 == jtor2

# Check a single linear solve
eq1.solve(profiles1)
eq2.solve(profiles2)

# Nonlinear solve
#freegs.solve(eq1, profiles, constrain)
#freegs.solve(eq2, profiles, constrain)

psi1 = eq1.psi()
psi2 = eq2.psi()

#psi1 = eq1.tokamak.psi(eq1.R, eq1.Z)
#psi2 = eq2.tokamak.psi(eq2.R, eq2.Z)

#psi1 = eq1.plasma_psi
#psi2 = eq2.plasma_psi

levels = linspace(amin(psi1), amax(psi1), 40)

plt.contour(eq1.R, eq1.Z, psi1, levels=levels, colors='k')
plt.contour(eq2.R, eq2.Z, psi2, levels=levels, colors='r')
plt.show()







import numpy as np

from . import jtor
from . import equilibrium


def test_psinorm_range():
    """Test that the profiles produce finite values outside core"""

    eq = equilibrium.Equilibrium(Rmin=0.5, Rmax=1.5, Zmin=-1.0, Zmax=1.0, nx=33, ny=33)

    for profiles in [
        jtor.ConstrainPaxisIp(eq, 1e3, 2e5, 2.0),
        jtor.ConstrainBetapIp(eq, 1.0, 2e5, 2.0),
    ]:
        current_density = profiles.Jtor(eq.R, eq.Z, eq.psi())
        assert np.all(np.isfinite(current_density))

        assert profiles.pprime(1.0) == 0.0
        assert profiles.pprime(1.1) == 0.0
        assert np.isfinite(profiles.pprime(-0.32))

        assert profiles.ffprime(1.0) == 0.0
        assert profiles.ffprime(1.1) == 0.0
        assert np.isfinite(profiles.ffprime(-0.32))



#!/usr/bin/env python

import freegs
import numpy as np
import matplotlib.pyplot as plt

start_resolution = 17
nrefinements = 5  # Number of refinements. Minimum 2
rtol = 1e-10  # Relative tolerance in Picard iteration
location = (1.2, 0.1)  # Location to record values at

############################################
# Generate low resolution solution

tokamak = freegs.machine.TestTokamak()

eq = freegs.Equilibrium(
    tokamak=tokamak,
    Rmin=0.1,
    Rmax=2.0,  # Radial domain
    Zmin=-1.0,
    Zmax=1.0,  # Height range
    nx=start_resolution,
    ny=start_resolution,  # Number of grid points
    boundary=freegs.boundary.freeBoundaryHagenow,
)

profiles = freegs.jtor.ConstrainPaxisIp(
    1e3,  # Plasma pressure on axis [Pascals]
    2e5,  # Plasma current [Amps]
    2.0,  # Vacuum f=R*Bt
)

xpoints = [(1.1, -0.6), (1.1, 0.8)]  # (R,Z) locations of X-points

isoflux = [(1.1, -0.6, 1.1, 0.6)]  # (R1,Z1, R2,Z2) pair of locations

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

freegs.solve(
    eq,  # The equilibrium to adjust
    profiles,  # The toroidal current profile function
    constrain,
    rtol=rtol,
    maxits=120,
)

############################################
# Now have initial solution

resolutions = [eq.R.shape[0]]
psivals = [eq.psiRZ(*location)]
brvals = [eq.Br(*location)]
volumevals = [eq.plasmaVolume()]
coilcurrents = [eq.tokamak["P1L"].current]

# List of l2 and l∞ norms
l2vals = []
linfvals = []

for i in range(nrefinements):

    print("\n========== Refining ===========\n")

    psi_old = eq.psi()
    R_old = eq.R
    Z_old = eq.Z

    # Increase resolution
    eq = freegs.equilibrium.refine(eq)

    # Re-solve
    freegs.solve(eq, profiles, constrain, rtol=rtol, maxits=120)

    resolutions.append(eq.R.shape[0])

    # Get the new psi on the old points
    psi_new = eq.psiRZ(R_old, Z_old)

    # Global norms of the change in psi
    l2 = np.sqrt(np.mean((psi_new - psi_old) ** 2))
    linf = np.amax(np.abs(psi_new - psi_old))

    l2vals.append(l2)
    linfvals.append(linf)

    # Point-wise values
    psivals.append(eq.psiRZ(*location))
    brvals.append(eq.Br(*location))

    # Coil current
    coilcurrents.append(eq.tokamak["P1L"].current)

    # Global quantity
    volumevals.append(eq.plasmaVolume())


resolutions = np.squeeze(np.array(resolutions))
psivals = np.squeeze(np.array(psivals))
brvals = np.squeeze(np.array(brvals))
volumevals = np.squeeze(np.array(volumevals))
coilcurrents = np.squeeze(np.array(coilcurrents))

l2vals = np.array(l2vals)
linfvals = np.array(linfvals)

fig, axes = plt.subplots(2, 3, sharex=True)


def plot_convergence(axis, title, values=None, diffs=None):
    # Absolute differences
    if diffs is None:
        diffs = np.abs(values[1:] - values[0:-1])

    # Convergence order. Note one shorter than diffs array
    orders = np.log(diffs[:-1] / diffs[1:]) / np.log(2.0)

    axis.plot(resolutions[1:] - 1, diffs)
    axis.set_title(title)
    axis.set_yscale("log")
    axis.set_xscale("log")

    for x, y, order in zip(resolutions[2:], diffs[1:], orders):
        axis.text(x, y, "{:.1f}".format(order))

    resolution_list = list(resolutions[1:] - 1)
    axis.set_xticks(resolution_list)
    axis.set_xticklabels(["{:d}".format(r) for r in resolution_list])
    axis.set_xticks([], minor=True)


plot_convergence(axes[0, 0], r"Change in $\psi$ at {}".format(location), values=psivals)
plot_convergence(axes[0, 1], r"$\psi$ difference $l_2$ norm", diffs=l2vals)
plot_convergence(axes[0, 2], r"$\psi$ difference $l_\infty$ norm", diffs=linfvals)
plot_convergence(axes[1, 0], "Br at {}".format(location), values=brvals)
plot_convergence(axes[1, 1], "Plasma volume", values=volumevals)
plot_convergence(axes[1, 2], "P1L coil current", values=coilcurrents)

plt.savefig("test-convergence.pdf")
plt.savefig("test-convergence.png")

plt.show()



#!/usr/bin/env python
#
# Example demonstrating functions for creating and finding X-points

import freegs

# Plotting routines
from freegs.plotting import plotEquilibrium, plotCoils, plotConstraints

import matplotlib.pyplot as plt

tokamak = freegs.machine.TestTokamak()
eq = freegs.Equilibrium(tokamak=tokamak, nx=256,ny=256)

##########################################################
# Calculate currents in coils to create X-points
# in specified locations
# 

xpoints = [(1.1, -0.8),   # (R,Z) locations of X-points
           (1.1, 0.8)]

control = freegs.control.constrain(xpoints=xpoints)
control(eq)  # Apply control to Equilibrium eq

psi = eq.psi()

print("=> Solved coil currents, created X-points")

ax = plotEquilibrium(eq, show=False)
plotCoils(tokamak.coils, axis=ax)
plotConstraints(control, axis=ax)
plt.show()

##########################################################
# Find critical points (O- and X-points)
# 
# 

import freegs.critical as critical
opt, xpt = critical.find_critical(eq.R, eq.Z, psi)

print("=> Found O- and X-points")

ax = plotEquilibrium(eq, show=False, oxpoints=False)
for r,z,_ in xpt:
    ax.plot(r,z,'ro')
for r,z,_ in opt:
    ax.plot(r,z,'go')
psi_bndry = xpt[0][2]
sep_contour = ax.contour(eq.R, eq.Z,psi, levels=[psi_bndry], colors='r')
psi_bndry = eq.psi_bndry
sep_contour = ax.contour(eq.R, eq.Z,psi, levels=[psi_bndry], colors='k')
plt.show()

##########################################################
# Create a mask array, 1 in the core and 0 outside
# 
# 

mask = critical.core_mask(eq.R, eq.Z, psi, opt, xpt)

print("=> Created X-point mask")

plt.contourf(eq.R, eq.Z, mask)
plt.show()



#!/usr/bin/env python

import freegs
from freegs.plotting import plotConstraints

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.MASTU_simple()


eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-2.1, Zmax=2.1,   # Height range
                        nx=65, ny=65)          # Number of grid points

#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        6e4, # Plasma pressure on axis [Pascals]
                                        1e6, # Plasma current [Amps]
                                        0.65, # vacuum f = R*Bt
                                        alpha_m = 1.0,
                                        alpha_n = 2.0)

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

Rx = 0.509
Zx = 1.291

Rmid = 1.34   # Outboard midplane
Rin = 0.3581  # Inboard midplane

xpoints = [(Rx, -Zx),   # (R,Z) locations of X-points
           (Rx,  Zx)]

isoflux = [(Rx,-Zx, Rmid, 0.0)   # Outboard midplane, lower X-point
           ,(Rx,Zx, Rmid, 0.0)   # Outboard midplane, upper X-point

           # Link inner and outer midplane locations
           ,(Rmid, 0.0, Rin, 0.0)

           # Separatrix in the divertor chamber
           ,(Rx,-Zx, 0.95, -1.77)
           ,(Rx, Zx, 0.95,  1.77)
           ]

constrain = freegs.control.constrain(xpoints=xpoints, gamma=8e-6, isoflux=isoflux)

constrain(eq)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The plasma profiles
             constrain,   # Plasma control constraints
             show=True)   # Shows results at each nonlinear iteration

#########################################
# Now adjust the equilibrium manually
# 

isoflux = [(Rx,-Zx, Rmid, 0.0)   # Outboard midplane, lower X-point
           ,(Rx,Zx, Rmid, 0.0)   # Outboard midplane, upper X-point

           ,(Rmid, 0.0, Rin, 0.0)
           
           ,(Rx,-Zx, 0.95, -1.77)
           ,(Rx, Zx, 0.95,  1.77)

           ,(Rx,-Zx, 0.76, -1.58)
           ,(Rx, Zx, 0.76,  1.58)
           
           ,(Rx,-Zx, 1.25, -1.8)
           ,(Rx, Zx, 1.25,  1.8)

           ]

constrain = freegs.control.constrain(xpoints=xpoints, gamma=1e-12, isoflux=isoflux)

# Turn off feedback control for all coils
for label, coil in tokamak.coils:
    coil.control = False

# Centre column coil
tokamak["Pc"].current = -4e4
    
# Turn on vertical feedback control
tokamak["P6"].control = True

# Coil in the "nose" of the divertor
tokamak["Dp"].current = -1000.0

# At top of divertor chamber
tokamak["D6"].current = 500.0
tokamak["D7"].current = 500.0

# X-point location
tokamak["Px"].current = 2000.0

tokamak["D1"].current = 1000.0

# Coil in outer corner 
tokamak["D5"].current = 1500.

# Coil at bottom centre
tokamak["D3"].current = 2800

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The plasma profiles
             constrain,   # Plasma control constraints
             show=True)   # Shows results at each nonlinear iteration

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq.plasmaCurrent()))
print("Pressure on axis: %e Pascals" % (eq.pressure(0.0)))
print("Poloidal beta: %e" % (eq.poloidalBeta()))
print("Plasma volume: %e m^3" % (eq.plasmaVolume()))

eq.tokamak.printCurrents()

axis = eq.plot(show=False)
eq.tokamak.plot(axis=axis)
constrain.plot(axis=axis)

##############################################
# Save to geqdsk file

from freegs import geqdsk

with open("mast-upgrade.geqdsk", "w") as f:
    geqdsk.write(eq, f)

# Call matplotlib show so plot pauses
import matplotlib.pyplot as plt
plt.show()



#!/usr/bin/env python

import freegs
import freegs.critical as critical
from freegs import geqdsk
from freegs.plotting import plotEquilibrium

import matplotlib.pyplot as plt
import numpy as np

'''
First creates a limited plasma and solves, before
creating a diverted plasma (which will have a larger CSA)
and solves under identical constraints. Due to the reduced CSA
of the limited plasma, one expects the plasma profiles to have a larger
magnitude when limited such that J is larger. J is expected to be larger
such that when J is integrated over a smaller CSA, the resultant Ip (and BetaP)
are the same. Consequently, different coil currents are also expected.
'''

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.TestTokamakLimited()

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-1.0, Zmax=1.0,   # Height range
                        nx=65, ny=65,          # Number of grid points
                        boundary=freegs.boundary.freeBoundaryHagenow)  # Boundary condition


#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainBetapIp(eq,
                                        0.15, # Plasma poloidal beta
                                        2e5, # Plasma current [Amps]
                                        2.0) # Vacuum f=R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(1.1, -0.6),   # (R,Z) locations of X-points
           (1.1, 0.8)]

isoflux = [(1.1,-0.6, 1.1,0.6),(1.1,-0.6,0.732,0.0426)] # (R1,Z1, R2,Z2) pair of locations

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

#########################################
# Nonlinear solve

freegs.solve(eq,
            profiles, 
            constrain,
            show=True,
            check_limited = True,
            limit_it = 0)

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq.plasmaCurrent()))
print("Plasma pressure on axis: %e Pascals" % (eq.pressure(0.0)))
print("Poloidal beta: %e" % (eq.poloidalBeta()))

# Currents in the coils
tokamak.printCurrents()

##############################################
# Final plot of equilibrium

axis = eq.plot(show=False)
eq.tokamak.plot(axis=axis, show=False)
constrain.plot(axis=axis, show=True)

###############################################
# Check that the core is masked correctly

fig, ax = plt.subplots()

ax.contour(eq.R,eq.Z,eq.psiN(),levels=30,colors='b')
ax.contour(eq.R,eq.Z,eq.psiN(),levels=[1.0],colors='orange')
ax.plot(eq.tokamak.wall.R,eq.tokamak.wall.Z,color='k')
opt, xpt = critical.find_critical(eq.R,eq.Z,eq.psi())
isoflux = np.array(
    freegs.critical.find_separatrix(eq, ntheta=101, opoint=opt, xpoint=xpt, psi=eq.psi())
)
ind = np.argmin(isoflux[:, 1])
rbdry = np.roll(isoflux[:, 0][::-1], -ind)
rbdry = np.append(rbdry,rbdry[0])
zbdry = np.roll(isoflux[:, 1][::-1], -ind)
zbdry = np.append(zbdry, zbdry[0])
ax.plot(rbdry,zbdry,'rx')
ax.contourf(eq.R,eq.Z,eq.mask)
ax.set_aspect('equal')
ax.set_xlabel('R(m)')
ax.set_ylabel('Z(m)')
plt.show()

from freegs import geqdsk

with open("limited.geqdsk", "w") as f:
    geqdsk.write(eq, f)

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.TestTokamak()

eq2 = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-1.0, Zmax=1.0,   # Height range
                        nx=65, ny=65,          # Number of grid points
                        boundary=freegs.boundary.freeBoundaryHagenow)  # Boundary condition


#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainBetapIp(eq2,
                                        0.15, # Plasma poloidal beta
                                        2e5, # Plasma current [Amps]
                                        2.0) # Vacuum f=R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(1.1, -0.6),   # (R,Z) locations of X-points
           (1.1, 0.8)]

isoflux = [(1.1,-0.6, 1.1,0.6),(1.1,-0.6,0.732,0.0426)] # (R1,Z1, R2,Z2) pair of locations

#current_lims = [(-150000.0,140000.0),(0.0,65000.0),(-105000.0,0.0),(-60000.0,0.0)]
#total_current = 350000.0

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)#, current_lims=current_lims, max_total_current = total_current)

#########################################
# Nonlinear solve

freegs.solve(eq2,
            profiles, 
            constrain,
            show=True,
            check_limited = True)

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq2.plasmaCurrent()))
print("Plasma pressure on axis: %e Pascals" % (eq2.pressure(0.0)))
print("Poloidal beta: %e" % (eq2.poloidalBeta()))

# Currents in the coils
tokamak.printCurrents()

##############################################
# Final plot of equilibrium

axis = eq2.plot(show=False)
eq2.tokamak.plot(axis=axis, show=False)
constrain.plot(axis=axis, show=True)

plt.show()

###############################################
# Check that the core is masked correctly

fig, ax = plt.subplots()

ax.contour(eq2.R,eq2.Z,eq2.psiN(),levels=30,colors='b')
ax.contour(eq2.R,eq2.Z,eq2.psiN(),levels=[1.0],colors='orange')
ax.plot(eq2.tokamak.wall.R,eq2.tokamak.wall.Z,color='k')
opt, xpt = critical.find_critical(eq2.R,eq2.Z,eq2.psi())
mask = critical.core_mask(eq2.R,eq2.Z,eq2.psi(),opt,xpt,eq2.psi_bndry)
ax.contourf(eq2.R,eq2.Z,mask)
ax.set_aspect('equal')
ax.set_xlabel('R(m)')
ax.set_ylabel('Z(m)')
plt.show()
#################

# Compare plasma profiles between limited and diverted plasmas
psi_levels = np.linspace(0.0,1.0,100,endpoint=True)

fig, ax = plt.subplots()

ax.plot(psi_levels,eq.pprime(psi_levels),color='r',label='limited')
ax.plot(psi_levels,eq2.pprime(psi_levels),color='b',label='diverted')
ax.set_xlabel('psiN')
ax.set_ylabel('pprime')
ax.legend()
plt.show()

fig, ax = plt.subplots()
ax.plot(psi_levels,eq.pressure(psi_levels),color='r',label='limited')
ax.plot(psi_levels,eq2.pressure(psi_levels),color='b',label='diverted')
ax.set_xlabel('psiN')
ax.set_ylabel('p')
ax.legend()
plt.show()

fig, ax = plt.subplots()
ax.plot(psi_levels,eq.ffprime(psi_levels),color='r',label='limited')
ax.plot(psi_levels,eq2.ffprime(psi_levels),color='b',label='diverted')
ax.set_xlabel('psiN')
ax.set_ylabel('ffprime')
ax.legend()
plt.show()

fig, ax = plt.subplots()
ax.plot(psi_levels,eq.fpol(psi_levels),color='r',label='limited')
ax.plot(psi_levels,eq2.fpol(psi_levels),color='b',label='diverted')
ax.set_xlabel('psiN')
ax.set_ylabel('fpol')
ax.legend()
plt.show()

#########################
# Load in the geqdsk of the limited plasma
tokamak = freegs.machine.TestTokamakLimited()

with open("limited.geqdsk") as f:
    eq3 = geqdsk.read(f, tokamak, show=True)

# Plot equilibrium
plotEquilibrium(eq3)



"""
Define a class of coil which contains a uniform current density
over a shaped region.

License
-------

Copyright 2019 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from . import quadrature
from . import polygons
from .gradshafranov import Greens, GreensBr, GreensBz
from shapely.geometry import Polygon

import numpy as np

from .coil import Coil


class ShapedCoil(Coil):
    """
    Represents a coil with a specified shape

    public members
    --------------

    R, Z - Location of the point coil/Locations of coil filaments
    current - current in the coil(s) in Amps
    turns   - Number of turns if using point coils
    control - enable or disable control system
    area    - Cross-section area in m^2

    The total toroidal current carried by the coil block is current * turns
    """

    # A dtype for converting to Numpy array and storing in HDF5 files
    dtype = np.dtype(
        [
            (str("RZlen"), int),  # Length of the R and Z arrays
            (str("R"), "10f8"),  # Note: Up to 10 points
            (str("Z"), "10f8"),  # Note: Up to 10 points
            (str("current"), np.float64),
            (str("turns"), int),
            (str("control"), bool),
            (str("npoints"), int),
        ]
    )

    def __init__(self, shape, current=0.0, turns=1, control=True, npoints=6):
        """
        Inputs
        ------
        shape:
            Outline of the coil shape as a list of points ``[(r1,z1),
            (r2,z2), ...]``. Must have more than two points
        current:
            The current in the circuit. The total current is current * turns
        turns:
            Number of turns in point coil(s) block. Total block current is current * turns
        control:
            enable or disable control system
        npoints:
            Number of quadrature points per triangle. Valid choices: 1, 3, 6

        """
        assert len(shape) > 2

        # Find the geometric middle of the coil
        # The R,Z properties have accessor functions to handle modifications
        self._R_centre = sum(r for r, z in shape) / len(shape)
        self._Z_centre = sum(z for r, z in shape) / len(shape)

        self.current = current
        self.turns = turns
        self.control = control
        self._area = abs(polygons.area(shape))
        self.shape = shape

        # The quadrature points to be used
        self.npoints_per_triangle = npoints
        self._points = quadrature.polygon_quad(shape, n=npoints)

    def controlPsi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z) due to a unit current in the circuit
        """
        result = 0.0
        for R_fil, Z_fil, weight in self._points:
            result += Greens(R_fil, Z_fil, R, Z) * weight
        # Multiply by turns so that toroidal current is current * turns
        return result * self.turns

    def controlBr(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z) due to a unit current
        """
        result = 0.0
        for R_fil, Z_fil, weight in self._points:
            result += GreensBr(R_fil, Z_fil, R, Z) * weight
        return result * self.turns

    def controlBz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z) due to a unit current
        """
        result = 0.0
        for R_fil, Z_fil, weight in self._points:
            result += GreensBz(R_fil, Z_fil, R, Z) * weight
        return result * self.turns

    def inShape(self,polygon):
        Shaped_Coil = Polygon([shape for shape in self.shape])
        return (polygon.intersection(Shaped_Coil).area) / (self._area)

    def __repr__(self):
        return "ShapedCoil({0}, current={1:.1f}, turns={2}, control={3})".format(
            self.shape, self.current, self.turns, self.control
        )

    @property
    def R(self):
        """
        Major radius of the coil in m
        """
        return self._R_centre

    @R.setter
    def R(self, Rnew):
        # Need to shift all points
        Rshift = Rnew - self._R_centre
        self._points = [(r + Rshift, z, w) for r, z, w in self._points]
        self._R_centre = Rnew

    @property
    def Z(self):
        """
        Height of the coil in m
        """
        return self._Z_centre

    @Z.setter
    def Z(self, Znew):
        # Need to shift all points
        Zshift = Znew - self._Z_centre
        self._points = [(r, z + Zshift, w) for r, z, w in self._points]
        self._Z_centre = Znew

    @property
    def area(self):
        return self._area

    @area.setter
    def area(self, area):
        raise ValueError("Area of a ShapedCoil is fixed")

    def plot(self, axis=None, show=False):
        """
        Plot the coil shape, using axis if given
        """
        import matplotlib.pyplot as plt

        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(111)

        r = [r for r, z in self.shape]
        z = [z for r, z in self.shape]
        axis.fill(r, z, color="gray")
        axis.plot(r, z, color="black")

        # Quadrature points
        # rquad = [r for r,z,w in self._points]
        # zquad = [z for r,z,w in self._points]
        # axis.plot(rquad, zquad, 'ro')

        return axis

    def to_numpy_array(self):
        """
        Helper method for writing output
        """
        RZlen = len(self.shape)
        R = np.zeros(10)
        Z = np.zeros(10)
        R[:RZlen] = [R for R, Z in self.shape]
        Z[:RZlen] = [Z for R, Z in self.shape]

        return np.array(
            (
                RZlen,
                R,
                Z,
                self.current,
                self.turns,
                self.control,
                self.npoints_per_triangle,
            ),
            dtype=self.dtype,
        )

    @classmethod
    def from_numpy_array(cls, value):
        if value.dtype != cls.dtype:
            raise ValueError(
                "Can't create {this} from dtype: {got} (expected: {dtype})".format(
                    this=type(cls), got=value.dtype, dtype=cls.dtype
                )
            )
        RZlen = value["RZlen"]
        R = value["R"][:RZlen]
        Z = value["Z"][:RZlen]
        current = value["current"]
        turns = value["turns"]
        control = value["control"]
        npoints = value["npoints"]

        return ShapedCoil(
            list(zip(R, Z)),
            current=current,
            turns=turns,
            control=control,
            npoints=npoints,
        )

    def __eq__(self, other):
        return (
            np.allclose(self.shape, other.shape)
            and self.current == other.current
            and self.turns == other.turns
            and self.control == other.control
            and self.npoints_per_triangle == other.npoints_per_triangle
        )

    def __ne__(self, other):
        return not self == other



"""
Defines class to represent the equilibrium
state, including plasma and coil currents
"""

from typing import Optional

from numpy import pi, meshgrid, linspace, exp, array
import numpy as np
from scipy import interpolate
from scipy.integrate import romb, cumulative_trapezoid

from .boundary import fixedBoundary, freeBoundary
from . import critical

from . import polygons

# Operators which define the G-S equation
from .gradshafranov import mu0, GSsparse, GSsparse4thOrder

# Multigrid solver
from . import multigrid

from . import machine

import matplotlib.pyplot as plt


class Equilibrium:
    """
    Represents the equilibrium state, including
    plasma and coil currents

    Attributes
    ----------
    R[nx,ny]:
        Read-only
    Z[nx,ny]:
        Read-only
    Rmin, Rmax:
        Read-only
    Zmin, Zmax:
        Read-only
    tokamak:
        The coils and circuits
    _applyBoundary:
    _solver:
        Grad-Shafranov elliptic solver
    _profiles:
         An object which calculates the toroidal current
    _constraints:
        Control system which adjusts coil currents to meet constraints
        e.g. X-point location and flux values
    """

    def __init__(
        self,
        tokamak: machine.Machine = machine.EmptyTokamak(),
        Rmin: float = 0.1,
        Rmax: float = 2.0,
        Zmin: float = -1.0,
        Zmax: float = 1.0,
        nx: int = 65,
        ny: int = 65,
        boundary=freeBoundary,
        psi: Optional[float] = None,
        current: float = 0.0,
        order: int = 4,
        check_limited: bool = False,
    ):
        """Initialises a plasma equilibrium

        Parameters
        ----------
        tokamak:
            The set of coils and currents
        Rmin, Rmax:
            Range of major radius R [m]
        Zmin, Zmax:
            Range of height Z [m]
        nx:
            Resolution in R. This must be ``2^n + 1``
        ny:
            Resolution in Z. This must be ``2^m + 1``
        boundary:
            The boundary condition, either `freeBoundary` or `fixedBoundary`
        psi:
            Magnetic flux. If None, use concentric circular flux
            surfaces as starting guess
        current:
            Plasma current
        order:
            The order of the differential operators to use. Valid values are 2 or 4
        check_limited:
            Checks if the plasma is limited
        """

        self.tokamak = tokamak

        self._applyBoundary = boundary

        self.Rmin = Rmin
        self.Rmax = Rmax
        self.Zmin = Zmin
        self.Zmax = Zmax
        self.nx = nx
        self.ny = ny

        self.R_1D = linspace(Rmin, Rmax, nx)
        self.Z_1D = linspace(Zmin, Zmax, ny)
        self.R, self.Z = meshgrid(self.R_1D, self.Z_1D, indexing="ij")

        self.dR = self.R[1, 0] - self.R[0, 0]
        self.dZ = self.Z[0, 1] - self.Z[0, 0]

        self.check_limited = check_limited
        self.is_limited = False
        self.Rlim = None
        self.Zlim = None

        if psi is None:
            # Starting guess for psi
            xx, yy = meshgrid(linspace(0, 1, nx), linspace(0, 1, ny), indexing="ij")
            psi = exp(-((xx - 0.5) ** 2 + (yy - 0.5) ** 2) / 0.4**2)

            psi[0, :] = 0.0
            psi[:, 0] = 0.0
            psi[-1, :] = 0.0
            psi[:, -1] = 0.0

        # Calculate coil Greens functions. This is an optimisation,
        # used in self.psi() to speed up calculations
        self._pgreen = tokamak.createPsiGreens(self.R, self.Z)

        self._current = current  # Plasma current
        self.Jtor = None

        self._updatePlasmaPsi(psi)  # Needs to be after _pgreen

        # Create the solver
        if order == 2:
            generator = GSsparse(Rmin, Rmax, Zmin, Zmax)
        elif order == 4:
            generator = GSsparse4thOrder(Rmin, Rmax, Zmin, Zmax)
        else:
            raise ValueError(
                "Invalid choice of order ({}). Valid values are 2 or 4.".format(order)
            )
        self.order = order

        self._solver = multigrid.createVcycle(
            nx, ny, generator, nlevels=1, ncycle=1, niter=2, direct=True
        )

    def setSolverVcycle(self, nlevels=1, ncycle=1, niter=1, direct=True):
        """
        Creates a new linear solver, based on the multigrid code

        nlevels  - Number of resolution levels, including original
        ncycle   - The number of V cycles
        niter    - Number of linear solver (Jacobi) iterations per level
        direct   - Use a direct solver at the coarsest level?

        """
        generator = GSsparse(self.Rmin, self.Rmax, self.Zmin, self.Zmax)
        nx, ny = self.R.shape

        self._solver = multigrid.createVcycle(
            nx,
            ny,
            generator,
            nlevels=nlevels,
            ncycle=ncycle,
            niter=niter,
            direct=direct,
        )

    def setSolver(self, solver):
        """
        Sets the linear solver to use. The given object/function must have a __call__ method
        which takes two inputs

        solver(x, b)

        where x is the initial guess. This should solve Ax = b, returning the result.

        """
        self._solver = solver

    def callSolver(self, psi, rhs):
        """
        Calls the psi solver, passing the initial guess and RHS arrays

        Parameters
        ----------
        psi:
            Initial guess for the solution (used if iterative)
        rhs:

        Returns
        -------
        float:
            Solution psi

        """
        return self._solver(psi, rhs)

    def getMachine(self):
        """
        Returns the handle of the machine, including coils
        """
        return self.tokamak

    def plasmaCurrent(self):
        """
        Plasma current [Amps]
        """
        return self._current

    def plasmaVolume(self):
        """Calculate the volume of the plasma in m^3"""

        dR = self.R[1, 0] - self.R[0, 0]
        dZ = self.Z[0, 1] - self.Z[0, 0]

        # Volume element
        dV = 2.0 * pi * self.R * dR * dZ

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        # Integrate volume in 2D
        return romb(romb(dV))

    def plasmaBr(self, R, Z):
        """
        Radial magnetic field due to plasma
        Br = -1/R dpsi/dZ
        """
        return -self.psi_func(R, Z, dy=1, grid=False) / R

    def plasmaBz(self, R, Z):
        """
        Vertical magnetic field due to plasma
        Bz = (1/R) dpsi/dR
        """
        return self.psi_func(R, Z, dx=1, grid=False) / R

    def Br(self, R, Z):
        """
        Total radial magnetic field
        """
        return self.plasmaBr(R, Z) + self.tokamak.Br(R, Z)

    def Bz(self, R, Z):
        """
        Total vertical magnetic field
        """
        return self.plasmaBz(R, Z) + self.tokamak.Bz(R, Z)

    def Bpol(self, R, Z):
        """
        Total poloidal magnetic field
        """
        Br = self.Br(R, Z)
        Bz = self.Bz(R, Z)
        return np.sqrt(Br * Br + Bz * Bz)

    def Btor(self, R, Z):
        """
        Toroidal magnetic field
        """
        # Normalised psi
        psi_norm = (self.psiRZ(R, Z) - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Get f = R * Btor in the core. May be invalid outside the core
        fpol = self.fpol(psi_norm)

        if self.mask is not None:
            # Get the values of the core mask at the requested R,Z locations
            # This is 1 in the core, 0 outside
            mask = self.mask_func(R, Z, grid=False)
            fpol = fpol * mask + (1.0 - mask) * self.fvac()

        return fpol / R

    def Btot(self, R, Z):
        """
        Total magnetic field
        """
        Br = self.Br(R, Z)
        Bz = self.Bz(R, Z)
        Btor = self.Btor(R, Z)
        return np.sqrt(Br * Br + Bz * Bz + Btor * Btor)

    def psi(self):
        """
        Total poloidal flux (psi), including contribution from
        plasma and external coils.
        """
        # return self.plasma_psi + self.tokamak.psi(self.R, self.Z)
        return self.plasma_psi + self.tokamak.calcPsiFromGreens(self._pgreen)

    def psiN(self):
        """
        Total poloidal flux (psi), including contribution from
        plasma and external coils. Normalised such that psiN = 0 on
        the magnetic axis and 1 on the LCFS.
        """
        # return self.plasma_psi + self.tokamak.psi(self.R, self.Z)
        return (self.psi() - self.psi_axis) / (self.psi_bndry - self.psi_axis)

    def psiRZ(self, R, Z):
        """
        Return poloidal flux psi at given (R,Z) location
        """
        return self.psi_func(R, Z, grid=False) + self.tokamak.psi(R, Z)

    def psiNRZ(self, R, Z):
        """
        Return poloidal flux psi at given (R,Z) location. Normalised such
        that psiN = 0 on the magnetic axis and 1 on the LCFS.
        """
        return (self.psiRZ(R, Z) - self.psi_axis) / (self.psi_bndry - self.psi_axis)

    def fpol(self, psinorm):
        """
        Return f = R*Bt at specified values of normalised psi
        """
        return self._profiles.fpol(psinorm)

    def fvac(self):
        """
        Return vacuum f = R*Bt
        """
        return self._profiles.fvac()

    def q(self, psinorm=None, npsi=100):
        """
        Returns safety factor q at specified values of normalised psi

        psinorm is a scalar, list or array of floats betweem 0 and 1.

        >>> safety_factor = eq.q([0.2, 0.5, 0.9])

        If psinorm is None, then q on a uniform psi grid will be returned,
        along with the psi values

        >>> psinorm, q = eq.q()

        Note: psinorm = 0 is the magnetic axis, and psinorm = 1 is the separatrix.
              If you request a value close to either of these limits, an extrapolation
              based on a 1D grid of values from 0.01 to 0.99 will be used. This gives
              smooth and continuous q-profiles, but comes at an increased computational
              cost.
        """
        if psinorm is None:
            # An array which doesn't include psinorm = 0 or 1
            psinorm = linspace(1.0 / (npsi + 1), 1.0, npsi, endpoint=False)
            return psinorm, critical.find_safety(self, psinorm=psinorm)

        elif np.any((psinorm < 0.01) | (psinorm > 0.99)):
            psinorm_inner = np.linspace(0.01, 0.99, num=npsi)
            q_inner = critical.find_safety(self, psinorm=psinorm_inner)

            interp = interpolate.interp1d(
                psinorm_inner, q_inner, kind="quadratic", fill_value="extrapolate"
            )
            result = interp(psinorm)
        else:
            result = critical.find_safety(self, psinorm=psinorm)

        # Convert to a scalar if only one result
        if np.size(result) == 1:
            return float(result)
        return result

    def tor_flux(self, psi=None):
        """
        Calculates toroidal flux at specified values of poloidal flux.
        >>> q = drho/dpsi
        """
        psiN = (psi - self.psi_axis) / (self.psi_bndry - self.psi_axis)
        # Get safety factor of these flux surfaces
        qvals = self.q(psiN)

        # Integrate q wrt psi to get rho. rho = 0 @ psiN = 0
        result = cumulative_trapezoid(qvals, psi, initial=0.0) * (-1.0 / (2.0 * np.pi))

        # Convert to a scalar if only one result
        if len(result) == 1:
            return result[0]
        return result

    def rhotor(self, psi=None):
        """
        Calculates normalised toroidal flux at specified values of
        poloidal flux.

        >>> rhotor = sqrt ( tor_flux/max(tor_flux)).

        Maximum toroidal flux shoud be at LCFS.
        """

        torflux = self.tor_flux(psi)

        psi = np.linspace(self.psi_axis, self.psi_bndry, 101, endpoint=True)
        torflux_for_LCFS = self.tor_flux(psi)

        max_torflux = np.max(torflux_for_LCFS)

        result = np.sqrt(torflux / max_torflux)

        if len(result) == 1:
            return result[0]
        return result

    def pprime(self, psinorm):
        """
        Return p' at given normalised psi
        """
        return self._profiles.pprime(psinorm)

    def ffprime(self, psinorm):
        """
        Return ff' at given normalised psi
        """
        return self._profiles.ffprime(psinorm)

    def pressure(self, psinorm):
        """
        Returns plasma pressure at specified values of normalised psi
        """
        return self._profiles.pressure(psinorm)

    def separatrix(self, npoints=360):
        """
        Returns an array of npoints (R, Z) coordinates of the separatrix,
        equally spaced in geometric poloidal angle.
        """
        return array(critical.find_separatrix(self, ntheta=npoints, psi=self.psi()))[
            :, 0:2
        ]

    def solve(self, profiles, Jtor=None, psi=None, psi_bndry=None):
        """
        Calculate the plasma equilibrium given new profiles
        replacing the current equilibrium.

        This performs the linear Grad-Shafranov solve

        Parameters
        ----------
        profiles:
            An object describing the plasma profiles.
            At minimum this must have methods:

            - ``.Jtor(R, Z, psi, psi_bndry) -> [nx, ny]``
            - ``.pprime(psinorm)``
            - ``.ffprime(psinorm)``
            - ``.pressure(psinorm)``
            - ``.fpol(psinorm)``

        Jtor :
            If supplied, a 2D array specifying the toroidal current at each
            (R,Z) point.  If not supplied, Jtor is calculated from profiles by
            finding O,X-points
        psi_bndry:
            Poloidal flux to use as the separatrix (plasma boundary). If not
            given then X-point locations are used.
        """

        self._profiles = profiles
        self._updateBoundaryPsi()

        if Jtor is None:
            # Calculate toroidal current density
            if psi is None:
                psi = self.psi()
            Jtor = profiles.Jtor(self.R, self.Z, psi, psi_bndry=psi_bndry)

        # Set plasma boundary
        # Note that the Equilibrium is passed to the boundary function
        # since the boundary may need to run the G-S solver (von Hagenow's method)
        self._applyBoundary(self, Jtor, self.plasma_psi)

        # Right hand side of G-S equation
        rhs = -mu0 * self.R * Jtor

        # Copy boundary conditions
        rhs[0, :] = self.plasma_psi[0, :]
        rhs[:, 0] = self.plasma_psi[:, 0]
        rhs[-1, :] = self.plasma_psi[-1, :]
        rhs[:, -1] = self.plasma_psi[:, -1]

        # Call elliptic solver
        plasma_psi = self._solver(self.plasma_psi, rhs)

        self._updatePlasmaPsi(plasma_psi)

        # Update plasma current
        dR = self.R[1, 0] - self.R[0, 0]
        dZ = self.Z[0, 1] - self.Z[0, 0]
        self._current = romb(romb(Jtor)) * dR * dZ
        self.Jtor = Jtor

    def _updateBoundaryPsi(self, psi=None):
        """
        For an input psi the magnetic axis and boundary psi are identified along
        with the core mask.

        Various logical checks occur, depending on whether or not the user
        wishes to check if the plasma is limited or not, as well as whether
        or not any xpoints are present.
        """

        if psi is None:
            psi = self.psi()

        opt, xpt = critical.find_critical(self.R, self.Z, psi)

        psi = psi

        if opt:
            # Magnetic axis flux taken as primary o-point flux
            self.psi_axis = opt[0][2]
            """
            Several options depending on if user wishes to check
            if the plasma becomes limited.
            """

            # The user wishes to check if the plasma is limited
            if self.check_limited and self.tokamak.wall:
                # A wall has actually been provided, proceed with checking

                # Obtain flux on machine limit points
                Rlimit = self.tokamak.limit_points_R
                Zlimit = self.tokamak.limit_points_Z

                """
                If an xpoint is present (plasma is potentianlly diverted)
                then we must remove any limit points above/below the
                primary xpoint as the PFR may land on these points,
                which would break the algorithm (at present) for extracting the boundary
                flux if the plasma were to infact be limited. There is a more advanced
                version of this alogorithm that is more robust that will be
                added in the future.
                """

                if xpt:
                    limit_args = np.ravel(
                        np.argwhere(abs(Zlimit) < abs(0.75 * xpt[0][1]))
                    )
                    Rlimit = Rlimit[limit_args]
                    Zlimit = Zlimit[limit_args]

                # Obtain the flux psi at these limiter points
                R = np.asarray(self.R[:, 0])
                Z = np.asarray(self.Z[0, :])

                # psi is transposed due to how FreeGS meshgrids R,Z
                psi_2d = interpolate.interp2d(x=R, y=Z, z=psi.T)

                # Get psi at the limit points
                psi_limit_points = np.zeros(len(Rlimit))
                for i in range(len(Rlimit)):
                    psi_limit_points[i] = psi_2d(Rlimit[i], Zlimit[i])[0]

                # Get index of maximum psi value
                indMax = np.argmax(psi_limit_points)

                # Extract R,Z of the contact point
                self.Rlim = Rlimit[indMax]
                self.Zlim = Zlimit[indMax]

                # Obtain maximum psi
                self.psi_limit = psi_limit_points[indMax]

                # Check if any xpoints are present
                if xpt:
                    # Get flux from the primary xpoint
                    self.psi_xpt = xpt[0][2]

                    # Choose between diverted or limited flux
                    self.psi_bndry = max(self.psi_xpt, self.psi_limit)

                    if self.psi_bndry == self.psi_limit:
                        self.is_limited = True

                    else:
                        self.is_limited = False

                    # Mask the core
                    self.mask = critical.core_mask(
                        self.R, self.Z, psi, opt, xpt, self.psi_bndry
                    )

                    # Use interpolation to find if a point is in the core.
                    self.mask_func = interpolate.RectBivariateSpline(
                        self.R[:, 0], self.Z[0, :], self.mask
                    )

                else:
                    # No xpoints, therefore psi_bndry = psi_limit
                    self.psi_bndry = self.psi_limit
                    self.is_limited = True
                    self.mask = None

            else:
                # Either a wall was not provided or the user did not wish to
                # check if the plasma was limited
                if xpt:
                    self.psi_xpt = xpt[0][2]
                    self.psi_bndry = self.psi_xpt
                    self.mask = critical.core_mask(self.R, self.Z, psi, opt, xpt)

                    # Use interpolation to find if a point is in the core.
                    self.mask_func = interpolate.RectBivariateSpline(
                        self.R[:, 0], self.Z[0, :], self.mask
                    )
                elif self._applyBoundary is fixedBoundary:
                    # No X-points, but using fixed boundary
                    self.psi_bndry = psi[0, 0]  # Value of psi on the boundary
                    self.mask = None  # All points are in the core region
                else:
                    self.psi_bndry = None
                    self.mask = None

                self.is_limited = False

    def _updatePlasmaPsi(self, plasma_psi):
        """
        Sets the plasma psi data, updates spline interpolation coefficients.
        Also updates:

        self.mask        2D (R,Z) array which is 1 in the core, 0 outside
        self.psi_axis    Value of psi on the magnetic axis
        self.psi_bndry   Value of psi on plasma boundary
        """
        self.plasma_psi = plasma_psi

        # Update spline interpolation
        self.psi_func = interpolate.RectBivariateSpline(
            self.R[:, 0], self.Z[0, :], plasma_psi
        )

        # Update the plasma axis and boundary flux as well as mask
        self._updateBoundaryPsi()

    def plot(self, axis=None, show=True, oxpoints=True) -> plt.Axes:
        """
        Plot the equilibrium flux surfaces

        Parameters
        ----------
        axis :
            Specify the axis on which to plot
        show :
            Call matplotlib.pyplot.show() before returning
        oxpoints :
            Plot X points as red circles, O points as green circles

        """
        from .plotting import plotEquilibrium

        return plotEquilibrium(self, axis=axis, show=show, oxpoints=oxpoints)

    def getForces(self):
        """
        Calculate forces on the coils

        Returns a dictionary of coil label -> force
        """
        return self.tokamak.getForces(self)

    def printForces(self):
        """
        Prints a table of forces on coils
        """
        print("Forces on coils")

        def print_forces(forces, prefix=""):
            for label, force in forces.items():
                if isinstance(force, dict):
                    print(prefix + label + " (circuit)")
                    print_forces(force, prefix=prefix + "  ")
                else:
                    print(
                        prefix
                        + label
                        + " : R = {0:.2f} kN , Z = {1:.2f} kN".format(
                            force[0] * 1e-3, force[1] * 1e-3
                        )
                    )

        print_forces(self.getForces())

    def innerOuterSeparatrix(self, Z=0.0):
        """
        Locate R co ordinates of separatrix at both
        inboard and outboard poloidal midplane (Z = 0)
        """
        # Find the closest index to requested Z
        Zindex = np.argmin(abs(self.Z[0, :] - Z))

        # Normalise psi at this Z index
        psinorm = (self.psi()[:, Zindex] - self.psi_axis) / (
            self.psi_bndry - self.psi_axis
        )

        # Start from the magnetic axis
        Rindex_axis = np.argmin(abs(self.R[:, 0] - self.Rmagnetic()))

        # Inner separatrix
        # Get the maximum index where psi > 1 in the R index range from 0 to Rindex_axis
        outside_inds = np.argwhere(psinorm[:Rindex_axis] > 1.0)

        if outside_inds.size == 0:
            R_sep_in = self.Rmin
        else:
            Rindex_inner = np.amax(outside_inds)

            # Separatrix should now be between Rindex_inner and Rindex_inner+1
            # Linear interpolation
            R_sep_in = (
                self.R[Rindex_inner, Zindex] * (1.0 - psinorm[Rindex_inner + 1])
                + self.R[Rindex_inner + 1, Zindex] * (psinorm[Rindex_inner] - 1.0)
            ) / (psinorm[Rindex_inner] - psinorm[Rindex_inner + 1])

        # Outer separatrix
        # Find the minimum index where psi > 1
        outside_inds = np.argwhere(psinorm[Rindex_axis:] > 1.0)

        if outside_inds.size == 0:
            R_sep_out = self.Rmax
        else:
            Rindex_outer = np.amin(outside_inds) + Rindex_axis

            # Separatrix should now be between Rindex_outer-1 and Rindex_outer
            R_sep_out = (
                self.R[Rindex_outer, Zindex] * (1.0 - psinorm[Rindex_outer - 1])
                + self.R[Rindex_outer - 1, Zindex] * (psinorm[Rindex_outer] - 1.0)
            ) / (psinorm[Rindex_outer] - psinorm[Rindex_outer - 1])

        return R_sep_in, R_sep_out

    def intersectsWall(self):
        """Assess whether or not the core plasma touches the vessel
        walls. Returns True if it does intersect.
        """
        separatrix = self.separatrix()  # Array [:,2]
        wall = self.tokamak.wall  # Wall object with R and Z members (lists)

        return polygons.intersect(separatrix[:, 0], separatrix[:, 1], wall.R, wall.Z)

    def magneticAxis(self):
        """Returns the location of the magnetic axis as a list [R,Z,psi]"""
        opt, xpt = critical.find_critical(self.R, self.Z, self.psi())
        return opt[0]

    def Rmagnetic(self):
        """The major radius R of magnetic major radius"""
        return self.magneticAxis()[0]

    def Zmagnetic(self):
        """The height Z of magnetic axis"""
        return self.magneticAxis()[1]

    def geometricAxis(self, npoints=360):
        """Locates geometric axis, returning [R,Z]. First locates the
        extrema points in R of the LCFS, wherein P3 is at the IMP and
        P1 is at the OMP.

        R0 = R(P3) + 0.5*(R(P1)-R(P3))
        z0 = 0.5*(Z(P1)+Z(P3))
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]

        Rlcfs = np.array([i[0] for i in separatrix])
        Zlcfs = np.array([i[1] for i in separatrix])

        ind_P1 = np.argmax(Rlcfs)
        ind_P3 = np.argmin(Rlcfs)

        P1 = np.array([Rlcfs[ind_P1], Zlcfs[ind_P1]])
        P3 = np.array([Rlcfs[ind_P3], Zlcfs[ind_P3]])

        R0 = P3[0] + 0.5 * (P1[0] - P3[0])
        z0 = 0.5 * (P1[1] + P3[1])

        C = np.array([R0, z0])

        return C

    def Rgeometric(self, npoints=360):
        """Locates major radius R of the geometric major radius."""
        return self.geometricAxis(npoints=npoints)[0]

    def Zgeometric(self, npoints=360):
        """Locates the height z of the geometric axis."""
        return self.geometricAxis(npoints=npoints)[1]

    def minorRadius(self, npoints=360):
        """Calculates minor radius of the plasma, a. First locates the
        extrema points in R of the LCFS, wherein P3 is at the IMP and
        P1 is at the OMP.

        a = 0.5*(R(P1) - R(P3))
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]

        Rlcfs = np.array([i[0] for i in separatrix])

        R_P1 = np.max(Rlcfs)
        R_P3 = np.min(Rlcfs)

        return 0.5 * (R_P1 - R_P3)

    def aspectRatio(self, npoints=360):
        """Calculates the plasma aspect ratio.

        A = R0/a where R0 = major radius, a = minor radius.
        """
        return self.Rgeometric(npoints=npoints) / self.minorRadius(npoints=npoints)

    def inverseAspectRatio(self, npoints=360):
        """Calculates inverse of the plasma aspect ratio.

        epsilon = 1/A
        A = R0/a where R0 = major radius, a = minor radius.
        """
        return self.minorRadius(npoints=npoints) / self.Rgeometric(npoints=npoints)

    def elongation(self, npoints=360):
        """Calculates the elongation, kappa, of the plasma. A large number
        of points should be supplied such that any primary xpoint(s) on
        the LCFS are captured. The R,Z of the primary x-point is NOT
        itself included in the R,Z of the LCFS as the plasma may be limited.
        P2 is the point at the upper extent of the plasma, and P4 is the point
        at the lower extent of the plasma.

        kappa = (Z(P2) - Z(P4))/a
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]

        Zlcfs = np.array([i[1] for i in separatrix])

        Z_P2 = np.max(Zlcfs)
        Z_P4 = np.min(Zlcfs)

        a = self.minorRadius(npoints=npoints)

        return 0.5 * (Z_P2 - Z_P4) / a

    def elongationUpper(self, npoints=360):
        """Calculates the upper elongation, kappa_u, of the plasma. A large number
        of points should be supplied such that any primary xpoint(s) on
        the LCFS are captured. The R,Z of the primary x-point is NOT
        itself included in the R,Z of the LCFS as the plasma may be limited.
        P2 is the point at the upper extent of the plasma.

        kappa_u = (Z(P2) - z0)/a
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]

        Zlcfs = np.array([i[1] for i in separatrix])

        Z_P2 = np.max(Zlcfs)

        z0 = self.Zgeometric(npoints=npoints)
        a = self.minorRadius(npoints=npoints)

        return (Z_P2 - z0) / a

    def elongationLower(self, npoints=360):
        """Calculates the lower elongation, kappa_l, of the plasma. A large number
        of points should be supplied such that any primary xpoint(s) on
        the LCFS are captured. The R,Z of the primary x-point is NOT
        itself included in the R,Z of the LCFS as the plasma may be limited.
        P2 is the point at the upper extent of the plasma.

        kappa_u = (z0 - Z(P4))/a
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]

        Zlcfs = np.array([i[1] for i in separatrix])

        Z_P4 = np.min(Zlcfs)

        z0 = self.Zgeometric(npoints=npoints)
        a = self.minorRadius(npoints=npoints)

        return (z0 - Z_P4) / a

    def effectiveElongation(self, npoints=360):
        """Calculates plasma effective elongation using the plasma volume"""
        return self.plasmaVolume() / (
            2.0
            * np.pi
            * self.Rgeometric(npoints=npoints)
            * np.pi
            * self.minorRadius(npoints=npoints) ** 2
        )

    def triangularityUpper(self, npoints=360):
        """Calculates plasma upper triangularity, delta_u.
        P2 is the point at the upper extent of the plasma.

        tri_u = (R0 - R(P2))/a
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]
        Rlcfs = np.array([i[0] for i in separatrix])
        Zlcfs = np.array([i[1] for i in separatrix])

        ind_P2 = np.argmax(Zlcfs)

        R_P2 = Rlcfs[ind_P2]

        R0 = self.Rgeometric(npoints=npoints)
        a = self.minorRadius(npoints=npoints)

        return (R0 - R_P2) / a

    def triangularityLower(self, npoints=360):
        """Calculates plasma upper triangularity, delta_u.
        P4 is the point at the lower extent of the plasma.

        tri_l = (R0 - R(P4))/a
        """

        # Get points along the LCFS
        separatrix = self.separatrix(npoints=npoints)  # Array [:,2]
        Rlcfs = np.array([i[0] for i in separatrix])
        Zlcfs = np.array([i[1] for i in separatrix])

        ind_P2 = np.argmax(Zlcfs)

        R_P2 = Rlcfs[ind_P2]

        R0 = self.Rgeometric(npoints=npoints)
        a = self.minorRadius(npoints=npoints)

        return (R0 - R_P2) / a

    def triangularity(self, npoints=360):
        """Calculates plasma triangularity, delta.

        Here delta is defined as the average of the upper
        and lower triangularities.
        """

        tri_u = self.triangularityUpper(npoints=npoints)
        tri_l = self.triangularityLower(npoints=npoints)

        return 0.5 * (tri_u + tri_l)

    def shafranovShift(self, npoints=360):
        """Calculates the plasma shafranov shift
        [delta_shafR,delta_shafZ] where

        delta_shafR = Rmagnetic - Rgeo
        delta_shafR = Zmagnetic - z0
        """

        Rmag = self.Rmagnetic()
        Zmag = self.Zmagnetic()

        Rgeo = self.Rgeometric()
        z0 = self.Zgeometric()

        return np.array([Rmag - Rgeo, Zmag - z0])

    def internalInductance1(self, npoints=360):
        """Calculates li1 plasma internal inductance"""

        R = self.R
        Z = self.Z
        # Produce array of Bpol^2 in (R,Z)
        B_polvals_2 = self.Bz(R, Z) ** 2 + self.Br(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        Ip = self.plasmaCurrent()
        R_geo = self.Rgeometric(npoints=npoints)
        elon = self.elongation(npoints=npoints)
        effective_elon = self.effectiveElongation(npoints=npoints)

        integral = romb(romb(B_polvals_2 * dV))
        return ((2 * integral) / ((mu0 * Ip) ** 2 * R_geo)) * (
            (1 + elon * elon) / (2.0 * effective_elon)
        )

    def internalInductance2(self):
        """Calculates li2 plasma internal inductance"""

        R = self.R
        Z = self.Z
        # Produce array of Bpol in (R,Z)
        B_polvals_2 = self.Br(R, Z) ** 2 + self.Bz(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ
        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        Ip = self.plasmaCurrent()
        R_mag = self.Rmagnetic()

        integral = romb(romb(B_polvals_2 * dV))
        return 2 * integral / ((mu0 * Ip) ** 2 * R_mag)

    def internalInductance3(self, npoints=360):
        """Calculates li3 plasma internal inductance"""

        R = self.R
        Z = self.Z
        # Produce array of Bpol in (R,Z)
        B_polvals_2 = self.Br(R, Z) ** 2 + self.Bz(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        Ip = self.plasmaCurrent()
        R_geo = self.Rgeometric(npoints=npoints)

        integral = romb(romb(B_polvals_2 * dV))
        return 2 * integral / ((mu0 * Ip) ** 2 * R_geo)

    def internalInductance(self, npoints=360):
        """Calculates plasma internal inductance li

        li = 4/(mu0*R0*Ip^2) * int(2piR*(Bp^2/2mu0)*dR*dZ)
           = 2/(mu0^2*R0*Ip^2)*int(2piR*Bp^2*dR*dZ)
        """

        R = self.R
        Z = self.Z
        # Produce array of Bpol in (R,Z)
        B_polvals_2 = self.Br(R, Z) ** 2 + self.Bz(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        Ip = self.plasmaCurrent()
        R_geo = self.Rgeometric(npoints=npoints)

        integral = romb(romb(B_polvals_2 * dV))
        return 2 * integral / (mu0 * mu0 * R_geo * Ip * Ip)

    def poloidalBeta(self):
        """Calculate plasma poloidal beta by integrating the thermal pressure
        and poloidal magnetic field pressure over the plasma volume."""

        R = self.R
        Z = self.Z

        # Produce array of Bpol in (R,Z)
        B_polvals_2 = self.Br(R, Z) ** 2 + self.Bz(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        # Normalised psi
        psi_norm = self.psiN()

        # Plasma pressure
        pressure = self.pressure(psi_norm)

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        pressure_integral = romb(romb(pressure * dV))
        field_integral_pol = romb(romb(B_polvals_2 * dV))
        return 2 * mu0 * pressure_integral / field_integral_pol

    def poloidalBeta2(self):
        """Return the poloidal beta
        betap = (8pi/mu0) * int(p)dRdZ / Ip^2
        """

        dR = self.R[1, 0] - self.R[0, 0]
        dZ = self.Z[0, 1] - self.Z[0, 0]

        # Normalised psi
        psi_norm = (self.psi() - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Plasma pressure
        pressure = self.pressure(psi_norm)
        if self.mask is not None:
            # If there is a masking function (X-points, limiters)
            pressure *= self.mask

        # Integrate pressure in 2D
        return (
            ((8.0 * pi) / mu0)
            * romb(romb(pressure))
            * dR
            * dZ
            / (self.plasmaCurrent() ** 2)
        )

    def poloidalBeta3(self):
        """Calculates alterantive poloidal beta definition."""

        R = self.R
        Z = self.Z

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        # Normalised psi
        psi_norm = (self.psi() - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Plasma pressure
        pressure = self.pressure(psi_norm)

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        pressure_integral = romb(romb(pressure * dV))
        Ip = self.plasmaCurrent()
        vol = self.plasmaVolume()
        r0 = self.Rgeometric()
        return 4 * vol * pressure_integral / (mu0 * Ip * Ip * r0)

    def toroidalBeta(self):
        """Calculate plasma toroidal beta by integrating the thermal pressure
        and toroidal magnetic field pressure over the plasma volume."""

        R = self.R
        Z = self.Z

        # Produce array of Btor in (R,Z)
        B_torvals_2 = self.Btor(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        # Normalised psi
        psi_norm = (self.psi() - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Plasma pressure
        pressure = self.pressure(psi_norm)

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        pressure_integral = romb(romb(pressure * dV))

        # Correct for errors in Btor and core masking
        np.nan_to_num(B_torvals_2, copy=False)

        field_integral_tor = romb(romb(B_torvals_2 * dV))
        return 2 * mu0 * pressure_integral / field_integral_tor

    def totalBeta(self):
        """Calculate plasma total beta"""
        return 1.0 / ((1.0 / self.poloidalBeta()) + (1.0 / self.toroidalBeta()))

    def betaN(self, npoints=360):
        """Calculate normalised plasma beta"""
        geo = self.geometricAxis()
        Bt = self.Btor(geo[0], geo[1])
        return (
            100.0
            * 1.0e06
            * self.toroidalBeta()
            * ((self.minorRadius() * Bt) / (self.plasmaCurrent()))
        )

    def pressure_ave(self):
        """Calculate average pressure, Pa."""

        R = self.R
        Z = self.Z

        # Produce array of Btor in (R,Z)
        B_torvals_2 = self.Btor(R, Z) ** 2

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        # Normalised psi
        psi_norm = (self.psi() - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Plasma pressure
        pressure = self.pressure(psi_norm)

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        pressure_integral = romb(romb(pressure * dV))
        plasmaVolume = romb(romb(dV))

        return pressure_integral / plasmaVolume

    def w_th(self):
        """
        Stored thermal energy in plasma, J.
        """

        R = self.R
        Z = self.Z

        dR = R[1, 0] - R[0, 0]
        dZ = Z[0, 1] - Z[0, 0]
        dV = 2.0 * np.pi * R * dR * dZ

        # Normalised psi
        psi_norm = (self.psi() - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Plasma pressure
        pressure = self.pressure(psi_norm)

        if self.mask is not None:  # Only include points in the core
            dV *= self.mask

        pressure_integral = romb(romb(pressure * dV))
        thermal_energy = (3.0 / 2.0) * pressure_integral

        return thermal_energy

    def qcyl(self):
        """
        Cylindrical safety factor.
        """

        eps = self.inverseAspectRatio()
        a = self.minorRadius()

        btor = self.fvac() / self.Rgeometric()
        Ip = self.plasmaCurrent()

        kappa = self.elongation()

        val = 0.5 * (1 + kappa * kappa) * ((2.0 * np.pi * a * eps * btor) / (mu0 * Ip))

        return val


def refine(eq, nx=None, ny=None):
    """
    Double grid resolution, returning a new equilibrium


    """
    # Interpolate the plasma psi
    # plasma_psi = multigrid.interpolate(eq.plasma_psi)
    # nx, ny = plasma_psi.shape

    # By default double the number of intervals
    if not nx:
        nx = 2 * (eq.R.shape[0] - 1) + 1
    if not ny:
        ny = 2 * (eq.R.shape[1] - 1) + 1

    result = Equilibrium(
        tokamak=eq.tokamak,
        Rmin=eq.Rmin,
        Rmax=eq.Rmax,
        Zmin=eq.Zmin,
        Zmax=eq.Zmax,
        boundary=eq._applyBoundary,
        order=eq.order,
        nx=nx,
        ny=ny,
    )

    plasma_psi = eq.psi_func(result.R, result.Z, grid=False)

    result._updatePlasmaPsi(plasma_psi)

    if hasattr(eq, "_profiles"):
        result._profiles = eq._profiles

    if hasattr(eq, "control"):
        result.control = eq.control

    return result


def coarsen(eq):
    """
    Reduce grid resolution, returning a new equilibrium
    """
    plasma_psi = multigrid.restrict(eq.plasma_psi)
    nx, ny = plasma_psi.shape

    result = Equilibrium(
        tokamak=eq.tokamak,
        Rmin=eq.Rmin,
        Rmax=eq.Rmax,
        Zmin=eq.Zmin,
        Zmax=eq.Zmax,
        nx=nx,
        ny=ny,
    )

    result._updatePlasmaPsi(plasma_psi)

    if hasattr(eq, "_profiles"):
        result._profiles = eq._profiles

    if hasattr(eq, "control"):
        result.control = eq.control

    return result


def newDomain(eq, Rmin=None, Rmax=None, Zmin=None, Zmax=None, nx=None, ny=None):
    """Creates a new Equilibrium, solving in a different domain.
    The domain size (Rmin, Rmax, Zmin, Zmax) and resolution (nx,ny)
    are taken from the input equilibrium eq if not specified.
    """
    if Rmin is None:
        Rmin = eq.Rmin
    if Rmax is None:
        Rmax = eq.Rmax
    if Zmin is None:
        Zmin = eq.Zmin
    if Zmax is None:
        Zmax = eq.Zmax
    if nx is None:
        nx = eq.R.shape[0]
    if ny is None:
        ny = eq.R.shape[0]

    # Create a new equilibrium with the new domain
    result = Equilibrium(
        tokamak=eq.tokamak, Rmin=Rmin, Rmax=Rmax, Zmin=Zmin, Zmax=Zmax, nx=nx, ny=ny
    )

    # Calculate the current on the old grid
    profiles = eq._profiles
    Jtor = profiles.Jtor(eq.R, eq.Z, eq.psi(), eq.psi_bndry)

    # Interpolate Jtor onto new grid
    Jtor_func = interpolate.RectBivariateSpline(eq.R[:, 0], eq.Z[0, :], Jtor)
    Jtor_new = Jtor_func(result.R, result.Z, grid=False)

    result._applyBoundary(result, Jtor_new, result.plasma_psi)

    # Right hand side of G-S equation
    rhs = -mu0 * result.R * Jtor_new

    # Copy boundary conditions
    rhs[0, :] = result.plasma_psi[0, :]
    rhs[:, 0] = result.plasma_psi[:, 0]
    rhs[-1, :] = result.plasma_psi[-1, :]
    rhs[:, -1] = result.plasma_psi[:, -1]

    # Call elliptic solver
    plasma_psi = result._solver(result.plasma_psi, rhs)

    result._updatePlasmaPsi(plasma_psi)

    # Solve once more, calculating Jtor using new psi
    result.solve(profiles)

    return result


if __name__ == "__main__":
    # Test the different spline interpolation routines

    from numpy import ravel
    import matplotlib.pyplot as plt

    import machine

    tokamak = machine.TestTokamak()

    Rmin = 0.1
    Rmax = 2.0

    eq = Equilibrium(tokamak, Rmin=Rmin, Rmax=Rmax)

    import constraints

    xpoints = [(1.2, -0.8), (1.2, 0.8)]
    constraints.xpointConstrain(eq, xpoints)

    psi = eq.psi()

    tck = interpolate.bisplrep(ravel(eq.R), ravel(eq.Z), ravel(psi))
    spline = interpolate.RectBivariateSpline(eq.R[:, 0], eq.Z[0, :], psi)
    f = interpolate.interp2d(eq.R[:, 0], eq.Z[0, :], psi, kind="cubic")

    plt.plot(eq.R[:, 10], psi[:, 10], "o")

    r = linspace(Rmin, Rmax, 1000)
    z = eq.Z[0, 10]
    plt.plot(r, f(r, z), label="f")

    plt.plot(r, spline(r, z), label="spline")

    plt.plot(r, interpolate.bisplev(r, z, tck), label="bisplev")

    plt.legend()
    plt.show()



"""
Plasma control system

Use constraints to adjust coil currents
"""

from numpy import dot, transpose, eye, array, inf
from numpy.linalg import inv, norm
import numpy as np
from scipy import optimize
from . import critical


class constrain(object):
    """
    Adjust coil currents using constraints. To use this class,
    first create an instance by specifying the constraints

    >>> controlsystem = constrain(xpoints = [(1.0, 1.1), (1.0,-1.0)])

    controlsystem will now attempt to create x-points at
    (R,Z) = (1.0, 1.1) and (1.0, -1.1) in any Equilibrium

    >>> controlsystem(eq)

    where eq is an Equilibrium object which is modified by
    the call.

    The constraints which can be set are:

    xpoints - A list of X-point (R,Z) locations

    isoflux - A list of tuples (R1,Z1, R2,Z2)

    psivals - A list of (R,Z,psi) values

    At least one of the above constraints must be included.

    gamma - A scalar, minimises the magnitude of the coil currents

    The following constraitns are entirely optional:

    current_lims - A list of tuples [(l1,u1),(l2,u2)...(lN,uN)] for the upper
    and lower bounds on the currents in each coil.

    max_total_current - The maximum total current through the coilset.
    """

    def __init__(
        self,
        xpoints=[],
        gamma=1e-12,
        isoflux=[],
        psivals=[],
        current_lims=None,
        max_total_current=None,
    ):
        """
        Create an instance, specifying the constraints to apply
        """
        self.xpoints = xpoints
        self.gamma = gamma
        self.isoflux = isoflux
        self.psivals = psivals
        self.current_lims = current_lims
        self.max_total_current = max_total_current

    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """

        tokamak = eq.getMachine()

        constraint_matrix = []
        constraint_rhs = []
        for xpt in self.xpoints:
            # Each x-point introduces two constraints
            # 1) Br = 0

            Br = eq.Br(xpt[0], xpt[1])

            # Add currents to cancel out this field
            constraint_rhs.append(-Br)
            constraint_matrix.append(tokamak.controlBr(xpt[0], xpt[1]))

            # 2) Bz = 0

            Bz = eq.Bz(xpt[0], xpt[1])

            # Add currents to cancel out this field
            constraint_rhs.append(-Bz)
            constraint_matrix.append(tokamak.controlBz(xpt[0], xpt[1]))

        # Constrain points to have the same flux
        for r1, z1, r2, z2 in self.isoflux:
            # Get Psi at (r1,z1) and (r2,z2)
            p1 = eq.psiRZ(r1, z1)
            p2 = eq.psiRZ(r2, z2)
            constraint_rhs.append(p2 - p1)

            # Coil responses
            c1 = tokamak.controlPsi(r1, z1)
            c2 = tokamak.controlPsi(r2, z2)
            # Control for the difference between p1 and p2
            c = [c1val - c2val for c1val, c2val in zip(c1, c2)]
            constraint_matrix.append(c)

        # Constrain the value of psi
        for r, z, psi in self.psivals:
            p1 = eq.psiRZ(r, z)
            constraint_rhs.append(psi - p1)

            # Coil responses
            c = tokamak.controlPsi(r, z)
            constraint_matrix.append(c)

        if not constraint_rhs:
            raise ValueError("No constraints given")

        # Constraint matrix
        A = array(constraint_matrix)
        b = np.reshape(array(constraint_rhs), (-1,))

        # Number of controls (length of x)
        ncontrols = A.shape[1]

        # First solve analytically by Tikhonov regularisation
        # minimise || Ax - b ||^2 + ||gamma x ||^2

        # Calculate the change in coil current
        self.current_change = dot(
            inv(dot(transpose(A), A) + self.gamma ** 2 * eye(ncontrols)),
            dot(transpose(A), b),
        )

        # Now use the initial analytical soln to guide constrained solve

        # Establish constraints on changes in coil currents from the present
        # and max/min coil current constraints

        current_change_bounds = []

        if self.current_lims is None:
            for i in range(ncontrols):
                current_change_bounds.append((-inf, inf))
        else:
            assert len(self.current_lims) == len(tokamak.coils), \
                f"Should provide current limits for all coils. Provided {len(self.current_lims)} limits for {len(tokamak.coils)} coils."
            for i, (_, coil) in enumerate(tokamak.coils):
                if coil.control:
                    lower_lim = self.current_lims[i][0] - coil.current
                    upper_lim = self.current_lims[i][1] - coil.current
                    current_change_bounds.append((lower_lim, upper_lim))

        current_change_bnds = array(current_change_bounds)

        # Reform the constraint matrices to include Tikhonov regularisation
        A2 = np.concatenate([A, self.gamma * eye(ncontrols)])
        b2 = np.concatenate([b, np.zeros(ncontrols)])

        # The objetive function to minimize
        # || A2x - b2 ||^2
        def objective(x):
            return (norm((A2 @ x) - b2)) ** 2

        # Additional constraints on the optimisation
        cons = []

        def max_total_currents(x):
            sum = 0.0
            for delta, i in zip(x, tokamak.controlCurrents()):
                sum += abs(delta + i)
            return -(sum - self.max_total_current)

        if self.max_total_current is not None:
            con1 = {"type": "ineq", "fun": max_total_currents}
            cons.append(con1)

        # Use the analytical current change as the initial guess
        if self.current_change.shape[0] > 0:
            x0 = self.current_change
            sol = optimize.minimize(
                objective, x0, method="SLSQP", bounds=current_change_bnds, constraints=cons
            )

            self.current_change = sol.x
            tokamak.controlAdjust(self.current_change)

        # Store info for user
        self.current_change = self.current_change

        # Ensure that the last constraint used is set in the Equilibrium
        eq._constraints = self

    def plot(self, axis=None, show=True):
        """
        Plots constraints used for coil current control

        axis     - Specify the axis on which to plot
        show     - Call matplotlib.pyplot.show() before returning

        """
        from .plotting import plotConstraints

        return plotConstraints(self, axis=axis, show=show)


class ConstrainPsi2D(object):
    """
    Adjusts coil currents to minimise the square differences
    between psi[R,Z] and a target psi.

    Ignores constant offset differences between psi array
    """

    def __init__(self, target_psi, weights=None):
        """
        target_psi : 2D (R,Z) array
            Must be the same size as the equilibrium psi

        weights : float or 2D array of same size as target_psi
            Relative importance of each (R,Z) point in the fitting
            By default every point is equally weighted
            Set points to zero to ignore them in fitting.

        """
        if weights is None:
            weights = np.full(target_psi.shape, 1.0)

        # Remove the average so constant offsets are ignored
        self.target_psi = target_psi - np.average(target_psi, weights=weights)

        self.weights = weights

    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """

        tokamak = eq.getMachine()
        start_currents = tokamak.controlCurrents()

        end_currents, _ = optimize.leastsq(
            self.psi_difference, start_currents, args=(eq,)
        )

        tokamak.setControlCurrents(end_currents)

        # Ensure that the last constraint used is set in the Equilibrium
        eq._constraints = self

    def psi_difference(self, currents, eq):
        """
        Difference between psi from equilibrium with the given currents
        and the target psi
        """
        eq.getMachine().setControlCurrents(currents)
        psi = eq.psi()
        psi_av = np.average(psi, weights=self.weights)
        return (
            (psi - psi_av - self.target_psi) * self.weights
        ).ravel()  # flatten array


class ConstrainPsiNorm2D(object):
    """
    Adjusts coil currents to minimise the square differences
    between normalised psi[R,Z] and a target normalised psi.
    """

    def __init__(self, target_psinorm, weights=1.0):
        """
        target_psinorm : 2D (R,Z) array
            Must be the same size as the equilibrium psi

        weights : float or 2D array of same size as target_psinorm
            Relative importance of each (R,Z) point in the fitting
            By default every point is equally weighted
            Set points to zero to ignore them in fitting.

        """
        self.target_psinorm = target_psinorm
        self.weights = weights

    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """

        tokamak = eq.getMachine()
        start_currents = tokamak.controlCurrents()

        end_currents, _ = optimize.leastsq(
            self.psinorm_difference, start_currents, args=(eq,)
        )

        tokamak.setControlCurrents(end_currents)

        # Ensure that the last constraint used is set in the Equilibrium
        eq._constraints = self

    def psinorm_difference(self, currents, eq):
        """
        Difference between normalised psi from equilibrium with the given currents
        and the target psinorm
        """
        eq.getMachine().setControlCurrents(currents)
        psi = eq.psi()

        eq._updateBoundaryPsi(psi)
        psi_bndry = eq.psi_bndry
        psi_axis = eq.psi_axis

        # Calculate normalised psi.
        # 0 = magnetic axis
        # 1 = plasma boundary
        psi_norm = (psi - psi_axis) / (psi_bndry - psi_axis)

        return (
            (psi_norm - self.target_psinorm) * self.weights
        ).ravel()  # flatten array


class ConstrainPsi2DAdvanced(object):
    """
    Adjusts coil currents to minimise the square differences
    between psi[R,Z] and a target psi.

    Attempts to also constrain the coil currents as in the 'constrain' class.
    """

    def __init__(
        self, target_psi, weights=1.0, current_lims=None, max_total_current=None
    ):
        """
        target_psinorm : 2D (R,Z) array
            Must be the same size as the equilibrium psi

        weights : float or 2D array of same size as target_psinorm
            Relative importance of each (R,Z) point in the fitting
            By default every point is equally weighted
            Set points to zero to ignore them in fitting.

        current_bounds: List of tuples
            Optional list of tuples representing constraints on coil currents to be used
            when reconstructing the equilibrium from the geqdsk file.
            [(l1,u1),(l2,u2)...(lN,uN)]

        Create an instance, specifying the constraints to apply
        """

        self.current_lims = current_lims
        self.target_psi = target_psi
        self.weights = weights

    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """

        tokamak = eq.getMachine()
        start_currents = np.asarray(tokamak.controlCurrents())
        ncontrols = len(start_currents)

        # In order for the optimisation to work, the initial guess must be within the
        # bounds supplied. Hence, check start_currents and adjust accordingly to be within bounds
        for i in range(ncontrols):
            bnd_upper = max(self.current_lims[i])
            bnd_lower = min(self.current_lims[i])
            sc = start_currents[i]
            if not (bnd_lower <= sc <= bnd_upper):
                if sc < bnd_lower:
                    start_currents[i] = bnd_lower
                else:
                    start_currents[i] = bnd_upper

        current_bounds = []

        for i in range(ncontrols):
            if self.current_lims is None:
                current_bounds.append((-inf, inf))
            else:
                bnd_upper = max(self.current_lims[i])
                bnd_lower = min(self.current_lims[i])
                current_bounds.append((bnd_lower, bnd_upper))

        current_bnds = array(current_bounds)

        # Least squares optimisation of difference in target v achieved normalised psi
        # applied with bounds on coil currents
        end_currents = optimize.minimize(
            self.psi_difference,
            start_currents,
            method="L-BFGS-B",
            bounds=current_bnds,
            args=(eq,),
        ).x

        # Set the latest coil currents
        tokamak.setControlCurrents(end_currents)

        # Ensure that the last constraint used is set in the Equilibrium
        eq._constraints = self

    def psi_difference(self, currents, eq):
        """
        Sum of the squares of the differences between the achieved
        psi and the target psi.
        """
        eq.getMachine().setControlCurrents(currents)
        psi = eq.psi()
        eq._updateBoundaryPsi(psi)

        psi_av = np.average(psi, weights=self.weights)
        diff = (psi - psi_av - self.target_psi) * self.weights
        sum_square_diff = np.sum(diff * diff)

        return sum_square_diff


class ConstrainPsiNorm2DAdvanced(object):
    """
    Adjusts coil currents to minimise the square differences
    between normalised psi[R,Z] and a target normalised psi.

    Attempts to also constrain the coil currents as in the 'constrain' class.
    """

    def __init__(self, target_psinorm, weights=1.0, current_lims=None):
        """
        target_psinorm : 2D (R,Z) array
            Must be the same size as the equilibrium psi

        weights : float or 2D array of same size as target_psinorm
            Relative importance of each (R,Z) point in the fitting
            By default every point is equally weighted
            Set points to zero to ignore them in fitting.

        current_bounds: List of tuples
            Optional list of tuples representing constraints on coil currents to be used
            when reconstructing the equilibrium from the geqdsk file.
            [(l1,u1),(l2,u2)...(lN,uN)]

        Create an instance, specifying the constraints to apply
        """

        self.current_lims = current_lims
        self.target_psinorm = target_psinorm
        self.weights = weights

    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """

        tokamak = eq.getMachine()
        start_currents = np.asarray(tokamak.controlCurrents())
        ncontrols = len(start_currents)

        # In order for the optimisation to work, the initial guess must be within the
        # bounds supplied. Hence, check start_currents and adjust accordingly to be within bounds
        for i in range(ncontrols):
            bnd_upper = max(self.current_lims[i])
            bnd_lower = min(self.current_lims[i])
            sc = start_currents[i]
            if not (bnd_lower <= sc <= bnd_upper):
                if sc < bnd_lower:
                    start_currents[i] = bnd_lower
                else:
                    start_currents[i] = bnd_upper

        current_bounds = []

        for i in range(ncontrols):
            if self.current_lims is None:
                current_bounds.append((-inf, inf))
            else:
                bnd_upper = max(self.current_lims[i])
                bnd_lower = min(self.current_lims[i])
                current_bounds.append((bnd_lower, bnd_upper))

        current_bnds = array(current_bounds)

        # Least squares optimisation of difference in target v achieved normalised psi
        # applied with bounds on coil currents
        end_currents = optimize.minimize(
            self.psinorm_difference,
            start_currents,
            method="L-BFGS-B",
            bounds=current_bnds,
            args=(eq,),
        ).x

        # Set the latest coil currents
        tokamak.setControlCurrents(end_currents)

        # Ensure that the last constraint used is set in the Equilibrium
        eq._constraints = self

    def psinorm_difference(self, currents, eq):
        """
        Sum of the squares of the differences between the achieved normalised
        psi and the target normalised psi.
        """
        eq.getMachine().setControlCurrents(currents)
        psi = eq.psi()

        eq._updateBoundaryPsi(psi)
        psi_bndry = eq.psi_bndry
        psi_axis = eq.psi_axis

        # Calculate normalised psi.
        # 0 = magnetic axis
        # 1 = plasma boundary
        psi_norm = (psi - psi_axis) / (psi_bndry - psi_axis)
        diff = (psi_norm - self.target_psinorm) * self.weights
        sum_square_diff = np.sum(diff * diff)

        return sum_square_diff



#!/usr/bin/env python
#
# Calculate the equilibrium for a plasma surrounded by a metal wall
# This is done by creating a ring of coils, with feedback control setting
# the poloidal flux to zero at the location of each coil.

import freegs
import numpy as np

#########################################
# Create a circular metal wall by using a ring of coils and psi constraints

R0 = 1.0 # Middle of the circle
rwall = 0.5  # Radius of the circular wall

npoints = 200 # Number of points on the wall

# Poloidal angles
thetas = np.linspace(0, 2*np.pi, npoints, endpoint=False)

# Points on the wall
Rwalls = R0 + rwall * np.cos(thetas)
Zwalls = rwall * np.sin(thetas)

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

coils = [ ("wall_"+str(theta), freegs.machine.Coil(R, Z))
          for theta, R, Z in zip(thetas, Rwalls, Zwalls) ]

tokamak = freegs.machine.Machine(coils)

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-1.0, Zmax=1.0,   # Height range
                        nx=65, ny=65,          # Number of grid points
                        boundary=freegs.boundary.freeBoundaryHagenow)  # Boundary condition

#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        1e4, # Plasma pressure on axis [Pascals]
                                        1e6, # Plasma current [Amps]
                                        2.0) # Vacuum f=R*Bt

#########################################
# Coil current constraints
#

# Same location as the coils
psivals = [ (R, Z, 0.0) for R, Z in zip(Rwalls, Zwalls) ]

constrain = freegs.control.constrain(psivals=psivals)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The toroidal current profile function
             constrain,   # Constraint function to set coil currents
             psi_bndry=0.0)  # Because no X-points, specify the separatrix psi

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq.plasmaCurrent()))
print("Plasma pressure on axis: %e Pascals" % (eq.pressure(0.0)))

##############################################
# Save to G-EQDSK file

from freegs import geqdsk

with open("metal-wall.geqdsk", "w") as f:
    geqdsk.write(eq, f)

##############################################
# Final plot

axis = eq.plot(show=False)
constrain.plot(axis=axis, show=True)




from freegs import machine, equilibrium, control, jtor, boundary, solve

tokamak = machine.TestTokamakSensor()

eq = equilibrium.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,
                        Zmin=-1.0, Zmax=1.0,
                        nx=65, ny=65,
                        boundary=boundary.freeBoundaryHagenow)

profiles = jtor.ConstrainPaxisIp(eq, 1e3, 2e5, 2.0)

xpoints = [(1.1, -0.6), (1.1, 0.8)]
isoflux = [(1.1,-0.6, 1.1,0.6)]

constrain = control.constrain(xpoints=xpoints, isoflux=isoflux)

solve(eq, profiles, constrain, show=True)

tokamak.printCurrents()
tokamak.printMeasurements(eq=eq)



#!/usr/bin/env python

import freegs

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.TestTokamak()

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-1.0, Zmax=1.0,   # Height range
                        nx=65, ny=65,          # Number of grid points
                        boundary=freegs.boundary.freeBoundaryHagenow)  # Boundary condition


#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        1e3, # Plasma pressure on axis [Pascals]
                                        2e5, # Plasma current [Amps]
                                        2.0) # Vacuum f=R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(1.1, -0.6),   # (R,Z) locations of X-points
           (1.1, 0.8)]

isoflux = [(1.1,-0.6, 1.1,0.6)] # (R1,Z1, R2,Z2) pair of locations

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The toroidal current profile function
             constrain,
             show=True)   # Constraint function to set coil currents

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq.plasmaCurrent()))
print("Plasma pressure on axis: %e Pascals" % (eq.pressure(0.0)))
print("Poloidal beta: %e" % (eq.poloidalBeta()))

# Currents in the coils
tokamak.printCurrents()

# Forces on the coils
eq.printForces()

print("\nSafety factor:\n\tpsi \t q")
for psi in [0.01, 0.9, 0.95]:
    print("\t{:.2f}\t{:.2f}".format(psi, eq.q(psi)))

##############################################
# Save to G-EQDSK file

from freegs import geqdsk

with open("lsn.geqdsk", "w") as f:
    geqdsk.write(eq, f)

##############################################
# Final plot of equilibrium

axis = eq.plot(show=False)
eq.tokamak.plot(axis=axis, show=False)
constrain.plot(axis=axis, show=True)

# Safety factor

import matplotlib.pyplot as plt
plt.plot(*eq.q())
plt.xlabel(r"Normalised $\psi$")
plt.ylabel("Safety factor")
plt.show()



import freegs
import numpy as np
from scipy.special import ellipk, ellipe
from scipy.constants import mu_0

def test_offaxis_Bfield():
    """
    Using the analytical equation for off axis b field due to a single coil loop
    And comparing to result returned from running the sensors
    Note - although the equation is analytic, its evaluation requires the use
    of elliptic integrals, therefore cannot be said to be exact
    """
    r = 4
    z = 7
    theta = np.pi / 2
    Rcoil = 1.5
    zcoil = 1
    current = 10000
    coil = [('Coil', freegs.machine.Coil(Rcoil, zcoil, current=current))]
    sensor = [freegs.machine.PoloidalFieldSensor(r, z, theta)]
    tokamak = freegs.machine.Machine(coil, sensors=sensor)

    # first want to write a calc for the analytical B field (with elliptical integrals)
    # defining the constants
    B0 = (current * mu_0) / (
            2 * Rcoil)  # this is the field on axis centre of coil plane
    a = r / Rcoil
    b = (z - zcoil) / Rcoil
    c = (z - zcoil) / r
    Q = (1 + a) ** 2 + b ** 2
    k = ((4 * a) / Q) ** 0.5

    # calculating B components
    Bz = B0 * (1 / (np.pi * (Q ** 0.5))) * (ellipe(k ** 2) * (
            (1 - a ** 2 - b ** 2) / (Q - 4 * a)) + ellipk(k ** 2))
    Br = B0 * (c / (np.pi * (Q ** 0.5))) * (ellipe(k ** 2) * (
            (1 + a ** 2 + b ** 2) / (Q - 4 * a)) - ellipk(k ** 2))
    B = Bz * np.sin(theta) + Br * np.cos(theta)

    # secondly want to pull the magentic field i am calculating from the sensors
    for sensor in tokamak.sensors:
        if isinstance(sensor, freegs.machine.PoloidalFieldSensor):
            sensor.get_measure(tokamak, None)
            assert np.isclose(sensor.measurement, B, atol=1e-8)


def test_xpoint_field():
    """
    Runs a basic equilibrium with constraints, checks if the sensors measure
    the correct constraints value
    """
    tokamak = freegs.machine.TestTokamak()
    tokamak.sensors = [freegs.machine.PoloidalFieldSensor(1.1, -0.6, 0),
                       freegs.machine.PoloidalFieldSensor(1.1, 0.6, 0),
                       freegs.machine.PoloidalFieldSensor(1.1, -0.6, np.pi/2),
                       freegs.machine.PoloidalFieldSensor(1.1, 0.6, np.pi/2)
                       ]

    eq = freegs.Equilibrium(tokamak=tokamak,
                            Rmin=0.1, Rmax=2.0,  # Radial domain
                            Zmin=-1.0, Zmax=1.0,  # Height range
                            nx=65, ny=65,  # Number of grid points
                            boundary=freegs.boundary.freeBoundary)  # Boundary condition

    profiles = freegs.jtor.ConstrainPaxisIp(eq, 1e3,
                                            # Plasma pressure on axis [Pascals]
                                            2e5,  # Plasma current [Amps]
                                            2.0)  # Vacuum f=R*Bt

    xpoints = [(1.1, -0.6),  # (R,Z) locations of X-points
               (1.1, 0.6)]

    isoflux = [(1.1, -0.6, 1.1, 0.6)]  # (R1,Z1, R2,Z2) pair of locations

    constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

    constrain(eq)

    tokamak.takeMeasurements(eq)

    for sensor in tokamak.sensors:
        assert np.isclose(sensor.measurement, 0.0, atol=1e-5)


def test_iso_flux():
    """
    Similarly to earlier test, checks the 2 points of isoflux and ensures
    sensor meaurement confirms constraint
    """
    tokamak = freegs.machine.TestTokamak()
    tokamak.sensors = [freegs.machine.FluxLoopSensor(1.1, -0.6),
                       freegs.machine.FluxLoopSensor(1.1, 0.6)]

    eq = freegs.Equilibrium(tokamak=tokamak,
                            Rmin=0.1, Rmax=2.0,  # Radial domain
                            Zmin=-1.0, Zmax=1.0,  # Height range
                            nx=65, ny=65,  # Number of grid points
                            boundary=freegs.boundary.freeBoundary)  # Boundary condition

    profiles = freegs.jtor.ConstrainPaxisIp(eq, 1e3,
                                            # Plasma pressure on axis [Pascals]
                                            2e5,  # Plasma current [Amps]
                                            2.0)  # Vacuum f=R*Bt

    xpoints = [(1.1, -0.6),  # (R,Z) locations of X-points
               (1.1, 0.6)]

    isoflux = [(1.1, -0.6, 1.1, 0.6)]  # (R1,Z1, R2,Z2) pair of locations

    constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

    constrain(eq)

    tokamak.takeMeasurements(eq)

    assert np.isclose(tokamak.sensors[0].measurement, tokamak.sensors[1].measurement,
                        atol=1e-8)


def test_flux():
    """
    Testing analytical flux agaisnt the numerical sensor output for a single coil
    """
    r = 4
    z = 7
    theta = np.pi / 2
    Rcoil = 1.5
    zcoil = 1
    current = 100000
    coil = [('Coil', freegs.machine.Coil(Rcoil, zcoil, current=current))]
    sensor = [freegs.machine.FluxLoopSensor(r, z)]
    tokamak = freegs.machine.Machine(coil, sensors=sensor)

    # first want to write a calc for the analytical flux (with greens function)

    psi = freegs.gradshafranov.Greens(Rcoil, zcoil, r, z) * current

    for sensor in tokamak.sensors:
        if isinstance(sensor, freegs.machine.FluxLoopSensor):
            sensor.get_measure(tokamak, None)
            assert np.isclose(sensor.measurement, psi, atol=1e-8)


def test_rog_around_coil():
    """
    Test single Rog around a coil
    """

    Rcoil = 1.5
    zcoil = 1
    zcoil2 = -1
    current = 100000.0
    coil = [('Coil 1', freegs.machine.Coil(Rcoil, zcoil, current=current)),
            ('Coil 2', freegs.machine.Coil(Rcoil, zcoil2, current=current))]
    sensor = [freegs.machine.RogowskiSensor([1.45, 1.45, 1.55, 1.55],
                                            [0.95, 1.05, 1.05, 0.95])]
    tokamak = freegs.machine.Machine(coil, sensors=sensor)

    tokamak.takeMeasurements()

    assert np.isclose(tokamak.sensors[0].measurement, current, atol=1)


def test_rog_around_Shapedcoil():
    """
    Test single Rog around a coil
    """

    Rcoil = 1.5
    zcoil = 1
    current = 100000.0
    coil = [('Coil 1', freegs.machine.ShapedCoil(
        [(0.95, -0.1), (0.95, 0.1), (1.05, 0.1), (1.05, -0.1)],
        current=current))]
    sensor = [freegs.machine.RogowskiSensor([0.9, 0.9, 1.0, 1.0],
                                            [-0.2, 0.2, 0.2, -0.2])]
    tokamak = freegs.machine.Machine(coil, sensors=sensor)

    tokamak.takeMeasurements()
    for sensor in tokamak.sensors:
        assert np.isclose(sensor.measurement, current / 2, atol=1)


def test_rog_around_Filamentcoil():

    tokamak = freegs.machine.EmptyTokamak()
    total_current = 300

    tokamak.coils = [('FILCOIL1',freegs.machine.FilamentCoil((0.5,0.5,0.5),(0.9,1,1.1), current=total_current)),('FILCOIL2',freegs.machine.FilamentCoil((0.5,0.5,0.5),(-0.9,-1,-1.1), current=total_current))]
    tokamak.sensors = [freegs.machine.RogowskiSensor([0.4,0.4,0.6,0.6], [0.8,1.2,1.2,0.8], name='UPPERCOIL'),freegs.machine.RogowskiSensor([0.4,0.4,0.6,0.6], [0.95,1.05,1.05,0.95], name='UPPERFIL')]

    tokamak.takeMeasurements()

    assert tokamak.sensors[0].measurement == total_current and tokamak.sensors[1].measurement == total_current / 3


def test_rog_around_circuit():

    tokamak = freegs.machine.EmptyTokamak()
    circuit_current = 100
    npoints = 3

    tokamak.coils = [('CIRCUIT', freegs.machine.Circuit([('FILCOIL1',freegs.machine.FilamentCoil((0.5,0.5,0.5),(0.9,1,1.1)), npoints),('FILCOIL2',freegs.machine.FilamentCoil((0.5,0.5,0.5),(-0.9,-1,-1.1)), npoints)], current=circuit_current))]
    tokamak.sensors = [freegs.machine.RogowskiSensor([0.4,0.4,0.6,0.6], [0.8,1.2,1.2,0.8], name='UPPERCOIL'),freegs.machine.RogowskiSensor([0.4,0.4,0.6,0.6], [0.95,1.05,1.05,0.95], name='UPPERFIL')]

    tokamak.takeMeasurements()

    assert tokamak.sensors[0].measurement == circuit_current * npoints and tokamak.sensors[1].measurement == circuit_current


def test_rog_with_plasma():
    """
    Testing Rog around a plasma profile
    """
    tokamak = freegs.machine.EmptyTokamak()  #
    tokamak.sensors = [
        freegs.machine.RogowskiSensor([0.1, 0.1, 2, 2], [-1, 1, 1, -1])]

    eq = freegs.Equilibrium(tokamak=tokamak,
                            Rmin=0.1, Rmax=2.0,  # Radial domain
                            Zmin=-1.0, Zmax=1.0,  # Height range
                            nx=65, ny=65,  # Number of grid points
                            boundary=freegs.boundary.freeBoundary)  # Boundary condition

    plasmacurrent = 300000
    profiles = freegs.jtor.ConstrainPaxisIp(eq, 1e3,
                                            # Plasma pressure on axis [Pascals]
                                            plasmacurrent,
                                            # Plasma current [Amps]
                                            2.0)  # Vacuum f=R*Bt

    eq.solve(profiles)
    tokamak.takeMeasurements(eq)
    print(tokamak.sensors[0].measurement)
    assert np.isclose(tokamak.sensors[0].measurement, plasmacurrent,atol=1000)



# Field line tracing

from builtins import object

import numpy as np
from scipy.integrate import odeint
from scipy import interpolate

from . import critical


class FieldTracer(object):
    """A class for following magnetic field lines"""

    def __init__(self, eq):
        """
        Initialise a FieldTracer with an Equilibrium eq
        """
        self._eq = eq

        if eq.tokamak.wall:
            # Machine has a wall, used to define edges
            self.updateEvolving = self.wallDomain
        else:
            # No wall, so just use the domain
            self.updateEvolving = self.eqDomain

    def eqDomain(self, R, Z, evolving):
        """Update an array `evolving`, of the same shape as R and Z.
        Set all entries to zero which correspond to (R,Z) points outside
        the domain."""

        eps = 1e-2

        evolving[
            np.logical_or(
                np.logical_or((R < self._eq.Rmin + eps), (R > self._eq.Rmax - eps)),
                np.logical_or((Z < self._eq.Zmin + eps), (Z > self._eq.Zmax - eps)),
            )
        ] = 0.0
        return evolving

    def wallDomain(self, R, Z, evolving):
        """Updates an array `evolving`, of the same shape as R and Z.
        Set all entries to zero which correspond to (R,Z) points outside
        the domain.
        This is done by counting intersections with the wall"""

        Rwall = self._eq.tokamak.wall.R
        Zwall = self._eq.tokamak.wall.Z

        # Location of the middle of the domain
        Rmid = 0.5 * (self._eq.Rmin + self._eq.Rmax)
        Zmid = 0.5 * (self._eq.Zmin + self._eq.Zmax)

        # True if a point is outside the wall
        outside = np.full(R.shape, False)

        nwall = len(Rwall)
        for wi in range(nwall):
            # Find intersection between line from (Rmid, Zmid) to (R,Z)
            # with wall segment from (Rwall[wi], Zwall[wi]) to (Rwall[wi+1], Zwall[wi+1])
            wip = (wi + 1) % nwall

            a = R - Rmid
            b = Rwall[wip] - Rwall[wi]
            c = Z - Zmid
            d = Zwall[wip] - Zwall[wi]

            dr = Rwall[wip] - Rmid
            dz = Zwall[wip] - Zmid

            det = a * d - b * c

            # Note: Here expect divide-by-zero errors occasionally
            # but the resulting values won't be used
            alpha = (d * dr - b * dz) / det  # Location along line 1 [0,1]
            beta = (a * dz - c * dr) / det  # Location along line 2 [0,1]

            # If the lines cross, change outside <-> inside
            # Note: If det is small then lines are nearly parallel
            outside ^= (
                (np.abs(det) > 1e-6)
                & (alpha > 0.0)
                & (alpha < 1.0)
                & (beta > 0.0)
                & (beta < 1.0)
            )
        evolving[outside] = 0.0

    def fieldDirection(self, pos, toroidal_angle, evolving, backward):
        """
        Calculate the magnetic field direction at a given pos
        """
        position = pos.reshape((-1, 3))
        R = position[:, 0]
        Z = position[:, 1]
        # Length is position[:,2]

        # Calculate magnetic field components
        Br = self._eq.Br(R, Z)
        Bz = self._eq.Bz(R, Z)
        Btor = self._eq.Btor(R, Z)

        B = np.sqrt(Br ** 2 + Bz ** 2 + Btor ** 2)

        # Detect when the boundary has been reached
        self.updateEvolving(R, Z, evolving)

        # Common factor in all equations
        evolving_R_Bt = evolving * R / Btor

        # Rate of change of position with toroidal angle phi
        dRdphi = evolving_R_Bt * Br
        dZdphi = evolving_R_Bt * Bz
        dldphi = evolving_R_Bt * B

        if backward:
            # Reverse direction
            dRdphi *= -1.0
            dZdphi *= -1.0

        return np.column_stack((dRdphi, dZdphi, dldphi)).flatten()

    def follow(self, Rstart, Zstart, angles, rtol=None, backward=False):
        """Follow magnetic field lines from (Rstart, Zstart) locations
        to given toroidal angles.
        If backward is True then the field lines are followed in the
        -B direction"""

        Rstart = np.array(Rstart)
        Zstart = np.array(Zstart)

        array_shape = Rstart.shape
        assert Zstart.shape == array_shape

        evolving = np.ones(array_shape)

        # (R,Z,length) with length=0 initially
        position = np.column_stack((Rstart, Zstart, np.zeros(array_shape))).flatten()

        result = odeint(
            self.fieldDirection, position, angles, args=(evolving, backward), rtol=rtol
        )

        return result.reshape(angles.shape + array_shape + (3,))


class LineCoordinates:
    """Coordinates of a field line

    Attributes
    ----------
    R:
        Major radius [m]
    Z:
        Height [m]
    length:
        Field line length [m]

    All {R, Z, length} are NumPy arrays of the same shape
    """

    def __init__(self, R, Z, length):
        self.R = R
        self.Z = Z
        self.length = length


def traceFieldLines(eq, solwidth=0.03, nlines=10, nturns=50, npoints=200, axis=None):
    """Trace field lines from the outboard midplane

    Inputs
    ------
    eq:
        Equilibrium object
    solwidth:
        The width of the SOL in meters
    nlines:
        Number of field lines to follow
    nturns:
        Number of times around the tokamak to follow
    npoints:
        Maximum number of points per line. May hit a wall
    axis:
        Matplotlib figure Axis. If given, field lines are plotted on the axis

    Returns
    -------
    forward: LineCoordinates
        The forward field line coordinates
    backward: LineCoordinates
        The backward field line coordinates

    Example
    -------

        >>> forward, backward = traceFieldLines(eq)

    """
    ft = FieldTracer(eq)

    # Find the edge of the plasma
    psi = eq.psi()
    opoint, xpoint = critical.find_critical(eq.R, eq.Z, psi)

    r0, z0, psi_axis = opoint[0]
    psi_bndry = eq.psi_bndry

    # Find outboard midplane
    psifunc = interpolate.RectBivariateSpline(
        eq.R[:, 0], eq.Z[0, :], (psi - psi_axis) / (psi_bndry - psi_axis)
    )

    rmid, zmid = critical.find_psisurface(
        eq, psifunc, r0, z0, r0 + 10.0, z0, psival=1.0
    )

    # Starting locations, just outside the separatrix
    # Start first point a bit away from separatrix
    rstart = np.linspace(rmid + (solwidth / nlines), rmid + solwidth, nlines)
    zstart = np.full(nlines, zmid)

    angles = np.linspace(0.0, nturns * 2 * np.pi, npoints)

    # Follow field lines
    line_forward = ft.follow(rstart, zstart, angles)
    line_backward = ft.follow(rstart, zstart, angles, backward=True)

    forward = LineCoordinates(
        line_forward[:, :, 0], line_forward[:, :, 1], line_forward[:, :, 2]
    )

    backward = LineCoordinates(
        line_backward[:, :, 0], line_backward[:, :, 1], line_backward[:, :, 2]
    )

    if axis:
        # Plot field lines
        axis.plot(forward.R, forward.Z)
        axis.plot(backward.R, backward.Z)

        # Mark the starting location
        axis.plot(rstart, zstart, color="k", linewidth=2)

    return forward, backward



from freegs import geqdsk
from freegs import machine
from freegs.plotting import plotEquilibrium

# Reading MAST equilibrium, up-down symmetric coils
tokamak = machine.MAST()

#with open("g014220.00200") as f:
with open("mast.geqdsk") as f:
  eq = geqdsk.read(f, tokamak, show=True)

plotEquilibrium(eq)




import freegs
from freegs import Equilibrium, fieldtracer, solve
from freegs.boundary import freeBoundaryHagenow, fixedBoundary
from freegs.control import constrain
from freegs.machine import TestTokamak

from io import BytesIO

STANDARD_PARAMS = dict(Rmin=0.1, Rmax=2.0, Zmin=-1.0, Zmax=1.0, nx=17, ny=17)


def ConstrainPaxisIp(*args):
    """Backwards compatibility shim, as the signature changed in 0.7.0"""
    eq, paxis, Ip, fvac = args

    if freegs.__version__ >= "0.7.0":
        return freegs.jtor.ConstrainPaxisIp(eq, paxis, Ip, fvac)
    return freegs.jtor.ConstrainPaxisIp(paxis, Ip, fvac)


def time_freeboundary_solve():
    eq = Equilibrium(
        tokamak=TestTokamak(), boundary=freeBoundaryHagenow, **STANDARD_PARAMS
    )

    profiles = ConstrainPaxisIp(eq, 1e3, 2e5, 2.0)
    xpoints = [(1.1, -0.6), (1.1, 0.8)]
    isoflux = [(1.1, -0.6, 1.1, 0.6)]
    control = constrain(xpoints=xpoints, isoflux=isoflux)
    solve(eq, profiles, control)


def time_fixedboundary_solve():
    eq = Equilibrium(boundary=fixedBoundary, **STANDARD_PARAMS)
    profiles = ConstrainPaxisIp(eq, 1e3, 1e5, 1.0)
    solve(eq, profiles)


def time_find_xpoints():
    tokamak = TestTokamak()
    eq = Equilibrium(tokamak=tokamak, **STANDARD_PARAMS)
    control = constrain(xpoints=[(1.1, -0.8), (1.1, 0.8)])
    control(eq)
    _ = eq.psi()


def time_sensors():
    tokamak = freegs.machine.TestTokamakSensor()
    eq = freegs.equilibrium.Equilibrium(
        tokamak=tokamak, boundary=freeBoundaryHagenow, **STANDARD_PARAMS
    )

    profiles = ConstrainPaxisIp(eq, 1e3, 2e5, 2.0)
    control = constrain(
        xpoints=[(1.1, -0.6), (1.1, 0.8)], isoflux=[(1.1, -0.6, 1.1, 0.6)]
    )
    solve(eq, profiles, control)


class TimeEquilibrium:
    """Various benchmarks that can reuse the same equilbrium instance"""

    def setup_cache(self):
        tokamak = TestTokamak()
        eq = Equilibrium(
            tokamak=tokamak, boundary=freeBoundaryHagenow, **STANDARD_PARAMS
        )
        profiles = ConstrainPaxisIp(eq, 1e4, 1e6, 2.0)
        xpoints = [(1.1, -0.6), (1.1, 0.8)]
        isoflux = [(1.1, -0.6, 1.1, 0.6)]
        control = constrain(xpoints=xpoints, isoflux=isoflux)
        solve(eq, profiles, control, maxits=25, atol=1e-3, rtol=1e-1)

        # SuperLU solver can't be pickled, but we don't need it for
        # these benchmarks
        eq.setSolver(None)
        return eq

    def time_dump(self, eq):
        memory_file = BytesIO()

        with freegs.OutputFile(memory_file, "w") as f:
            f.write_equilibrium(eq)

        with freegs.OutputFile(memory_file, "r") as f:
            _ = f.read_equilibrium()

    def time_find_critical(self, eq):
        _ = freegs.critical.find_critical(eq.R, eq.Z, eq.psi())

    def time_refine(self, eq):
        freegs.equilibrium.refine(eq)

    def time_trace_fieldlines(self, eq):
        fieldtracer.traceFieldLines(eq, nturns=1, nlines=1, npoints=10)



"""
Classes and routines to represent coils and circuits

License
-------

Copyright 2022 Chris Marsden, Tokamak Energy. Email: chris.marsden@tokamakenergy.co.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from . import polygons
from shapely import geometry
from .coil import Coil, AreaCurrentLimit
from .gradshafranov import Greens, GreensBr, GreensBz


def populate_with_fils(shape, Nfils):
    """
    Takes the 2D cross section of a PF coil, via shape, and populates it
    (somewhat) uniformally with ~Nfils point sources representing filaments.
    """

    Rshape = np.array([i[0] for i in shape])
    Zshape = np.array([i[1] for i in shape])
    Rshape = np.append(Rshape, Rshape[0])
    Zshape = np.append(Zshape, Zshape[0])
    rmin = np.min(Rshape)
    rmax = np.max(Rshape)
    zmin = np.min(Zshape)
    zmax = np.max(Zshape)

    my_polygon = geometry.Polygon(shape)

    """
    How this works: Start by roughly estimating the d = dR, dZ spacing
    between the points, using the shape area and Nfils.
    """

    A = abs(polygons.area(shape))
    d = np.sqrt(A / Nfils)

    """
    Now we create a (staggered) grid across the bounding box containing
    the supplied shape. We then scan through each point on the grid and
    check if it lies inside the shape. We count the number of points that
    lie inside the shape. If this is less than Nfils, increase the grid
    resolution a little until eventually the number of points inside the grid
    => Nfils. Once this is obtained, linearly interpolate in dR,dZ to better
    estimate the d(=dR,dZ) that will give as close to Nfils points inside the shape
    as possible.
    """

    # Track d vs nInside
    d_data = []
    nInside_data = []

    nInside = 0

    while nInside < Nfils:

        nInside = 0
        i = 0  # Row counter used for staggering every other row
        # Create the (non-staggered) grid points for the given d(=dR,dZ)
        Rgrid = np.arange(rmin, rmax, d, dtype=float)
        Zgrid = np.arange(zmin, zmax, d, dtype=float)

        nR = len(Rgrid)
        nZ = len(Zgrid)

        for i in range(nZ):

            zpoint = Zgrid[i]

            if i % 2:  # Every other row, stagger R coords by 0.5*dR
                offset = 0.5 * d
            else:

                offset = 0.0

            for j in range(nR):

                rpoint = Rgrid[j] + offset

                point = geometry.Point(rpoint, zpoint)
                if my_polygon.contains(point):
                    nInside += 1

        # End of looping over points. Note d and nInside
        d_data.append(d)
        nInside_data.append(nInside)

        # Shrink spacing by 0.5%
        d *= 0.995

    # Finished and have got nInside >= Nfils
    d_data = np.asarray(d_data)
    nInside = np.asarray(nInside_data)

    # Now linearly interpolate to get the d that corresponds
    # to the closest nInside to Nfils
    d_opt = np.interp([Nfils * 1.1], nInside_data, d_data)

    # Now get the locations of the filaments for this d = d_opt where nInside ~ Nfils
    rfils = []
    zfils = []

    # Number of filaments landing inside
    nInside = 0

    i = 0  # Row counter used for staggering every other row
    # Create the (non-staggered) grid points for the given d(=dR,dZ)
    Rgrid = np.arange(rmin, rmax, d_opt, dtype=float)
    Zgrid = np.arange(zmin, zmax, d_opt, dtype=float)

    nR = len(Rgrid)
    nZ = len(Zgrid)

    for i in range(nZ):

        zpoint = Zgrid[i]

        if i % 2:  # Every other row, stagger R coords by 0.5*dR
            offset = 0.5 * d_opt
        else:

            offset = 0.0

        for j in range(nR):

            rpoint = Rgrid[j] + offset

            point = geometry.Point(rpoint, zpoint)

            if my_polygon.contains(point):
                nInside += 1
                rfils.append(rpoint)
                zfils.append(zpoint)

    rfils = np.asarray(rfils)
    zfils = np.asarray(zfils)

    return rfils, zfils


class FilamentCoil(Coil):
    """
    This class represents a coil broken down into multiple
    filaments, with each filament acting as a point source
    of current.

    Note: Filaments are wired in parallel, so a the current
    through a single turn is shared between the filaments.

    public members
    --------------

    R, Z    - Location of the coil
    current - current in each turn of the coil in Amps
    turns   - Number of turns
    control - enable or disable control system
    area    - Cross-section area in m^2

    The total toroidal current carried by the coil is current * turns
    """

    # A dtype for converting to Numpy array and storing in HDF5 files
    dtype = np.dtype(
        [
            (str("RZlen"), int),  # Length of R and Z arrays
            (str("R"), "500f8"),  # Up to 100 points
            (str("Z"), "500f8"),
            (str("current"), np.float64),
            (str("turns"), int),
            (str("control"), bool),
            (str("npoints"), int),
        ]
    )

    def __init__(
        self,
        Rfil=None,
        Zfil=None,
        shape=None,
        current=0.0,
        Nfils=50,
        turns=1,
        control=True,
    ):
        """
        Inputs
        ------
        shape:
            Outline of the coil shape as a list of points ``[(r1,z1),
            (r2,z2), ...]`` Must have more than two points. If not
            provided, plotting the coil may not be entirely accurate.
        Rfil, Zfil:
            Locations of coil filaments (lists/arrays). This is optional.
            If these are not provided then the filaments themselves are populated
            automatically across the cross-section of the coil.
        current :
            Current in each turn of the coil in Amps
        Nfils   :
            Number of filaments. Only used when Rfil is None.
        turns   :
            Number of turns in point coil(s) block. Total block current is current * turns
        control :
            Enable or disable control system.

        Note
        ----
        The number of filaments does not equal the number of turns;
        each turn of the coil might consist of multiple filaments.

        """

        if shape is None:
            # Note: NumPy min/max works with float where builtin doesn't
            minR = np.min(Rfil)
            maxR = np.max(Rfil)
            minZ = np.min(Zfil)
            maxZ = np.max(Zfil)

            shape = [
                (minR, minZ),
                (minR, maxZ),
                (maxR, maxZ),
                (maxR, minZ),
                (minR, minZ),
            ]

        assert len(shape) > 2

        # Find the geometric middle of the coil
        # The R,Z properties have accessor functions to handle modifications
        self._R_centre = sum(r for r, z in shape) / len(shape)
        self._Z_centre = sum(z for r, z in shape) / len(shape)

        self.current = current
        self.turns = turns
        self.control = control
        self._area = abs(polygons.area(shape))
        self.shape = shape

        if Rfil is None:
            # No filaments provided. Need to populate the coil cross
            # section with -Nfils filaments.
            Rfil, Zfil = populate_with_fils(self.shape, Nfils)

        Rfil = np.asarray(Rfil)
        Zfil = np.asarray(Zfil)
        if Rfil.ndim == 0:
            self.points = [(Rfil, Zfil)]
            self.npoints = 1
        else:
            self.points = np.array(list(zip(Rfil, Zfil)))
            self.npoints = len(Rfil)

    def controlPsi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z) due to a unit current

        Note: This is multiplied by current to get coil Psi
        """
        result = 0.0
        for R_fil, Z_fil in self.points:
            result += Greens(R_fil, Z_fil, R, Z)
        result = result * self.turns / float(self.npoints)
        return result

    def controlBr(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z) due to a unit current
        """
        result = 0.0
        for R_fil, Z_fil in self.points:
            result += GreensBr(R_fil, Z_fil, R, Z)
        result = result * self.turns / float(self.npoints)
        return result

    def controlBz(self, R, Z):
        """
        Calculate axial magnetic field Br at (R,Z) due to a unit current
        """
        result = 0.0
        for R_fil, Z_fil in self.points:
            result += GreensBz(R_fil, Z_fil, R, Z)
        result = result * self.turns / float(self.npoints)
        return result

    def inShape(self,polygon):
        counter = 0
        for r, z in self.points:
            if polygon.contains(geometry.Point(r, z)):
                counter += 1 / self.npoints
        return counter

    def __repr__(self):
        return "FilamentCoil({0}, current={1:.1f}, turns={2}, control={3})".format(
            self.points, self.current, self.turns, self.control
        )

    @property
    def R(self):
        """
        Major radius of the coil in m
        """
        return self._R_centre

    @R.setter
    def R(self, Rnew):
        # Need to shift all points
        Rshift = Rnew - self._R_centre
        self.points = [(r + Rshift, z) for r, z in self.points]
        self._R_centre = Rnew

    @property
    def Z(self):
        """
        Height of the coil in m
        """
        return self._Z_centre

    @Z.setter
    def Z(self, Znew):
        # Need to shift all points
        Zshift = Znew - self._Z_centre
        self.points = [(r, z + Zshift) for r, z in self.points]
        self._Z_centre = Znew

    @property
    def area(self):
        return self._area

    @area.setter
    def area(self, area):
        raise ValueError("Area of a FilamentCoil is fixed")

    def plot(self, axis=None, show=False):
        """
        Plot the coil points, using axis if given
        """
        import matplotlib.pyplot as plt

        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(111)

        r = [r for r, z in self.shape]
        z = [z for r, z in self.shape]
        axis.fill(r, z, color="gray")
        axis.plot(r, z, color="black")

        r = [r for r, z in self.points]
        z = [z for r, z in self.points]
        axis.plot(r, z, "kx")

        # Quadrature points
        # rquad = [r for r,z,w in self._points]
        # zquad = [z for r,z,w in self._points]
        # axis.plot(rquad, zquad, 'ro')

        return axis



#!/usr/bin/env python
#
# Grad-Shafranov solver example
# Fixed boundary (square domain) with no X-points
#

# Plasma equilibrium (Grad-Shafranov) solver
import freegs

# Boundary conditions
import freegs.boundary as boundary

eq = freegs.Equilibrium(Rmin=0.1, Rmax=2.0,
                        Zmin=-1.0, Zmax=1.0,
                        nx=65, ny=65,
                        boundary=boundary.fixedBoundary)

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        1e3, # Plasma pressure on axis [Pascals]
                                        1e5, # Plasma current [Amps]
                                        1.0) # fvac = R*Bt

# Nonlinear solver for Grad-Shafranov equation
freegs.solve(eq,           # The equilibrium to adjust
            profiles,	   # The toroidal current profile function
            show = True)

print("Done!")

# Some diagnostics
print("Poloidal beta: {}".format(eq.poloidalBeta()))
print("Pressure on axis: {} Pa".format(eq.pressure(0.0)))

# Plot equilibrium
from freegs.plotting import plotEquilibrium
plotEquilibrium(eq)





#!/usr/bin/env python

import freegs

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.MAST()

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-2.0, Zmax=2.0,   # Height range
                        nx=65, ny=65)          # Number of grid points

#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        3e3, # Plasma pressure on axis [Pascals]
                                        7e5, # Plasma current [Amps]
                                        0.4) # vacuum f = R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(0.7, -1.1),   # (R,Z) locations of X-points
           (0.7, 1.1)]

isoflux = [(0.7,-1.1, 1.45, 0.0)   # Outboard midplane, lower X-point
           ,(0.7,1.1, 1.45, 0.0)   # Outboard midplane, upper X-point
#           ,(0.7,-1.1, 1.5,-1.9)  # Lower X-point, lower outer leg
#           ,(0.7,1.1,  1.5, 1.9)  # Upper X-point, upper outer leg
           ]

constrain = freegs.control.constrain(xpoints=xpoints, gamma=1e-12, isoflux=isoflux)

constrain(eq)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The plasma profiles
             constrain,   # Plasma control constraints
             show=True)   # Shows results at each nonlinear iteration

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq.plasmaCurrent()))
print("Pressure on axis: %e Pascals" % (eq.pressure(0.0)))
print("Plasma poloidal beta: %e" % (eq.poloidalBeta()))
print("Plasma volume: %e m^3" % (eq.plasmaVolume()))

eq.tokamak.printCurrents()

##############################################
# Save to geqdsk file

from freegs import geqdsk

with open("mast.geqdsk", "w") as f:
    geqdsk.write(eq, f)

# Call matplotlib show so plot pauses
import matplotlib.pyplot as plt
plt.show()



import freegs

import io
from numpy import allclose


def test_readwrite():
    """Test reading/writing to a file round-trip"""

    for tokamak in [freegs.machine.TestTokamak(), freegs.machine.MAST_sym()]:

        eq = freegs.Equilibrium(
            tokamak=tokamak,
            Rmin=0.1,
            Rmax=2.0,
            Zmin=-1.0,
            Zmax=1.0,
            nx=17,
            ny=17,
            boundary=freegs.boundary.freeBoundaryHagenow,
        )
        profiles = freegs.jtor.ConstrainPaxisIp(eq, 1e4, 1e6, 2.0)

        # Note here the X-point locations and isoflux locations are not the same.
        # The result will be an unbalanced double null configuration, where the
        # X-points are on different flux surfaces.
        xpoints = [(1.1, -0.6), (1.1, 0.8)]
        isoflux = [(1.1, -0.6, 1.1, 0.6)]
        constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux)

        freegs.solve(eq, profiles, constrain, maxits=25, atol=1e-3, rtol=1e-1)

        memory_file = io.BytesIO()

        with freegs.OutputFile(memory_file, "w") as f:
            f.write_equilibrium(eq)

        with freegs.OutputFile(memory_file, "r") as f:
            read_eq = f.read_equilibrium()

        assert tokamak == read_eq.tokamak
        assert allclose(eq.psi(), read_eq.psi())



# Read MAST-U GEQDSK file, calculate connection length

from freegs import geqdsk
from freegs import machine

tokamak = machine.MASTU()

with open("mast-upgrade.geqdsk") as f:
    eq = geqdsk.read(f, tokamak, show=False)

from freegs import fieldtracer
import matplotlib.pyplot as plt

# Plot equilibrium
axis = eq.plot(show=False)

# Trace field lines both directions along the magnetic field
# By passing axis 
forward, backward = fieldtracer.traceFieldLines(eq, axis=axis, nturns=50)

plt.savefig("mast-upgrade-fieldtrace.pdf")
plt.show()

# Plot field line length from midplane to target
plt.plot(forward.R[0,:], forward.length[-1,:], label="Forward")
plt.plot(backward.R[0,:], backward.length[-1,:], label="Backward")
plt.legend()
plt.xlabel("Starting major radius [m]")
plt.ylabel("Parallel connection length [m]")

plt.savefig("mast-upgrade-lpar.pdf")
plt.show()



#!/usr/bin/env python

import freegs

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.TestTokamak()

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.0,    # Radial domain
                        Zmin=-1.0, Zmax=1.0,   # Height range
                        nx=65, ny=65,          # Number of grid points
                        boundary=freegs.boundary.freeBoundaryHagenow)  # Boundary condition


#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
										1e3, # Plasma pressure on axis [Pascals]
                                        2e5, # Plasma current [Amps]
                                        2.0) # Vacuum f=R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(1.1, -0.6),   # (R,Z) locations of X-points
           (1.1, 0.6)]

isoflux = [(1.1,-0.6, 1.1,0.6), # (R1,Z1, R2,Z2) pair of locations
           (1.7, 0.0, 0.84, 0.0)]

constrain = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux, gamma = 1e-17)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The toroidal current profile function
             constrain)   # Constraint function to set coil currents


            
# Currents in the coils
tokamak.printCurrents()

# Forces on the coils
eq.printForces()

############################
# Optimise
# Minimise the maximum force on the coils, while avoiding intersection of the LCFS and walls
# by modifying the radius of the P2U and P2L coils.

from freegs import optimise as opt

print("Starting optimisation")

best_eq = opt.optimise(eq,  # Starting equilibrium
                       # List of controls
                       [opt.CoilRadius("P2U"),
                        opt.CoilRadius("P2L"), opt.CoilHeight("P2L")],
                       # The function to minimise
                       opt.weighted_sum(opt.max_coil_force, opt.no_wall_intersection),
                       N=10,  # Number of solutions in each generation
                       maxgen=20, # How many generations
                       monitor=opt.PlotMonitor())  # Plot the best in each generation

print("Finished optimisation")

# Forces on the coils
best_eq.printForces()
best_eq.plot()

